# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import sys
import json
import yaml
import random
import hashlib
import subprocess
import numpy as np
import datetime as dt
from os.path import dirname, isdir, splitext

from metagenomix._io_utils import (
    mkdr, get_roundtrip, print_status_table, get_versions)


class CreateScripts(object):

    def __init__(self, config, databases, workflow, commands):
        self.config = config
        self.databases = databases
        self.commands = commands
        self.graph = workflow.graph
        self.cmd = []
        self.cmds = {}
        self.chunks = {}
        self.soft = ''
        self.sh = ''
        self.run = {'database': {}, 'software': {}}
        self.job_fps = []
        self.job_name = ''
        self.modules = {}
        self.pjct = self.get_prjct()
        self.scheduler = self.get_scheduler()
        self.time = dt.datetime.now().strftime("%d/%m/%Y-%H:%M")
        self.scripts = []
        self.hash = ''

    def make_dirs(self):
        for name, soft in self.commands.softs.items():
            for directory in sorted(soft.dirs):
                if not isdir(directory):
                    os.makedirs(directory)

    def get_scheduler(self):
        if self.config.jobs:
            if self.config.torque:
                return 'qsub'
            else:
                return 'sbatch'
        else:
            return 'sh'

    def get_prjct(self):
        prjct = [x for x in self.config.project if x.lower() not in 'aeiouy']
        if prjct:
            return ''.join(prjct)
        else:
            return self.config.project

    def get_chunks(self, n_chunks):
        self.chunks = {}
        if n_chunks and len(self.cmds) >= n_chunks:
            cmds_d = dict(enumerate(self.cmds))
            chunks = [list(x) for x in np.array_split(list(cmds_d), n_chunks)]
            for cdx, chunk in enumerate(chunks):
                if len(chunk):
                    self.chunks[str(cdx)] = [cmds_d[x] for x in chunk]
        else:
            for key in self.cmds:
                if isinstance(key, str):
                    self.chunks[key] = [key]
                else:
                    self.chunks['_'.join([k for k in key if k])] = [key]

    def get_main_sh(self, name, soft=None) -> str:
        if soft:
            main = '%s/run.sh' % soft.dir
            self.run['software'][name] = main
        else:
            main = '%s/run_%s.sh' % (self.sh.rsplit('/', 2)[0], name)
            self.run['database'][name] = main
        return main

    def write_main(self, name, soft=None, soft_hash=None) -> None:
        main = self.get_main_sh(name, soft)
        with open(main, 'w') as o:
            if len(self.job_fps):
                job_dir = dirname(self.job_fps[0])
                o.write('mkdir -p %s/output\n' % job_dir)
                o.write('cd %s/output\n' % job_dir)
                for jdx, job_fp in enumerate(self.job_fps):
                    o.write('%s %s\n' % (self.scheduler, job_fp))
                    if self.config.dev:
                        if not isdir('%s/output' % job_dir):
                            os.makedirs('%s/output' % job_dir)
                        f = 'slurm-%s_000000%s' % (self.job_name, jdx)
                        for e in ['.o', '.e']:
                            job_std = '%s/output/%s%s' % (job_dir, f, e)
                            with open(job_std, 'w') as o_dev:
                                for r in range(10, random.randrange(11, 100)):
                                    o_dev.write('random %s\n' % r)
                                o_dev.write('Done!\n')
                self.job_fps = []

    def get_provenance_fp(self, name, soft=None) -> str:
        if soft:
            fp = '%s/provenance.txt' % soft.dir
        else:
            fp = '%s/provenance_%s.txt' % (self.sh.rsplit('/', 2)[0], name)
        return fp

    def write_provenance(self, name, soft=None, soft_hash=None) -> None:
        steps = self.graph.paths[soft.name][0]
        step_max = max([len(x) for x in steps])
        provenance_fp = self.get_provenance_fp(name, soft)
        with open(provenance_fp, 'w') as o:
            o.write("Pipeline steps to this output (and analysis type):\n")
            for sdx, step in enumerate(steps):
                o.write('%s\t%s%s\t: %s\n' % (
                    sdx, step, (' ' * (step_max-len(step))),
                    self.config.tools.get(step, '')))
            o.write("\nParameters for these steps' outputs:\n")
            for sdx, step in enumerate(steps):
                if step not in self.commands.softs:
                    o.write('\n ===== %s: %s (no param) =====\n' % (sdx, step))
                    continue
                o.write('\n========== %s: %s ==========\n' % (sdx, step))
                params_show = self.commands.softs[step].params.copy()
                if 'databases' in params_show:
                    databases = params_show['databases']
                    del params_show['databases']
                    params = {'search': {soft.name.split('_')[-1]: params_show,
                                         'databases': databases}}
                    o.write('%s\n' % yaml.dump(params))
                else:
                    o.write('%s\n' % yaml.dump(params_show))

    def database_cmds(self):
        for db, cmds in self.databases.commands.items():
            self.cmds = cmds
            self.get_chunks(self.config.params['chunks'])
            self.write_jobs(db)
            self.write_main(db)

    def get_modules(self, name: str):
        self.modules = self.config.modules.get(name, [])

    def scratch(self, soft, key, cmds):
        if soft.params['scratch'] and self.config.jobs and key in soft.io:
            roundtrip = get_roundtrip(soft.io[key])
            scratch_cmds = ['\n# Move to SCRATCH_FOLDER']
            scratch_cmds += roundtrip['to']
            scratch_cmds += ['\n# %s commands (%s)' % (soft.name, key)]
            scratch_cmds += cmds
            if self.config.move_back:
                scratch_cmds += ['\n# Move from SCRATCH_FOLDER']
                scratch_cmds += roundtrip['from']
            self.cmds[key] = scratch_cmds
        else:
            self.cmds[key] = cmds

    def get_cmds(self, soft):
        self.cmds = {}
        for sam_or_pool, cmds in soft.cmds.items():
            if isinstance(cmds, list):
                self.scratch(soft, sam_or_pool, cmds)
            elif isinstance(cmds, dict):
                if sam_or_pool in self.commands.pools:
                    for group, group_cmds in cmds.items():
                        self.scratch(soft, (sam_or_pool, group), group_cmds)
                else:
                    self.scratch(soft, sam_or_pool, cmds)
            else:
                sys.exit('The collected commands are neither list of dict!')

    def print_status(self, m, sdx, name, soft):
        gap = (m - len(name) - len(str(sdx))) + 1
        print('\t%s [%s] %s%s' % (sdx, name, ('.' * gap), ('.' * 8)), end=' ')
        print_status_table(soft, self.config.show_status)

    def get_hash(self, hash_string):
        h = hashlib.blake2b(digest_size=10)
        h.update(hash_string.encode('utf-8'))
        hashed = str(h.hexdigest())
        return hashed

    def get_soft_hash(self, soft=None) -> str:
        hashed = None
        if soft:
            steps = self.graph.paths[soft.name]
            hash_string = ''.join([json.dumps(step) + json.dumps(soft.params)
                                   for step in steps])
            self.hash += hash_string
            hashed = self.get_hash(hash_string)
        return hashed

    def versioning(self):
        versions_dir = '%s/_versions' % self.config.dir
        if not isdir(versions_dir):
            os.makedirs(versions_dir)
        self.write_versions(versions_dir)

    def write_versions(self, versions_dir):
        hashed = self.get_hash(self.hash)
        versions_fp = versions_dir + '/version_hash-' + hashed + '.txt'
        versions = get_versions(versions_fp)
        with open(versions_fp, 'w') as o:
            for script in self.scripts:
                o.write('%s\n' % script)
            o.write('\n--------- versions ---------\n')
            for version in versions:
                o.write('Date: %s\n' % version)
            o.write('Date: %s\n' % self.time)
            o.write('\n----------------------------\n')
        print('\n[config version] Written: %s' % versions_fp)

    def software_cmds(self):
        m = max(len(x) for x in self.commands.softs) + 1
        for sdx, (name, soft) in enumerate(self.commands.softs.items()):
            self.print_status(m, sdx, name, soft)
            if not len(soft.cmds):
                continue
            self.get_modules(name)
            self.get_cmds(soft)
            self.get_chunks(soft.params['chunks'])
            self.write_jobs(name, soft)
            soft_hash = self.get_soft_hash(soft)
            self.write_main(name, soft, soft_hash)
            self.write_provenance(name, soft, soft_hash)

    def get_sh(self, name: str, chunk: str, soft=None) -> None:
        """

        Parameters
        ----------
        name : str
            Name of the current software of the pipeline workflow
        chunk : str
            Name of the current chunk of commands
        soft
        """
        if soft:
            self.sh = '%s/jobs/run_%s.sh' % (
                soft.dir, chunk.replace('/', '_'))
        else:
            self.sh = '%s/databases/jobs/build_%s_%s.sh' % (
                self.config.dir, name, chunk)
        mkdr(dirname(self.sh))

    def prep_script(self, params: dict) -> None:
        # mandatory options
        self.cmd = [
            'Xhpc',
            '-i', self.sh,
            '-j', self.job_name,
            '-t', str(params['time']),
            '-c', str(params['cpus']),
            '-M', str(params['mem']), params['mem_dim'],
            '--no-stat']
        # whether the cpus requests is per node
        if params['nodes']:
            self.cmd.extend(['-n', str(params['nodes'])])
        # always provide an account
        if self.config.account:
            self.cmd.extend(['-a', self.config.account])
        # get the job script file path and use it
        job_script = '%s.slm' % splitext(self.sh)[0]
        if self.config.torque:
            self.cmd.append('--torque')
            job_script = '%s.pbs' % splitext(self.sh)[0]
        self.job_fps.append(job_script)
        self.cmd.extend(['-o', job_script])
        # machine-specific setup: env activating and slurm partition
        if params['machine']:
            self.cmd.append('--%s' % params['machine'])
        if params['partition']:
            self.cmd.append('--p-partition %s' % params['partition'])
        # whether an environment must be used for the current software
        if not self.modules and params['env']:
            self.cmd.extend(['-e', params['env']])
        # setup the scratch location to be used for the current software
        if isinstance(params['scratch'], int):
            self.cmd.extend(['--localscratch', str(params['scratch'])])
        elif params['scratch'] == 'scratch':
            self.cmd.append('--scratch')
        elif params['scratch'] == 'userscratch':
            self.cmd.append('--userscratch')
        if not self.config.verbose:
            self.cmd.append('--quiet')

    def call_cmd(self):
        cmd = ' '.join(self.cmd)
        if self.config.verbose:
            print('[Running]', cmd)
        subprocess.call(cmd.split())
        os.remove(self.sh)

    def write_chunks(self, chunk_keys: list, soft):
        with open(self.sh, 'w') as sh:
            if self.modules:
                sh.write('module purge\n')
            for module in self.modules:
                sh.write('module load %s\n' % module)
            for chunk_key in chunk_keys:
                for cmd in self.cmds[chunk_key]:
                    if soft.params['scratch'] and self.config.jobs:
                        sh.write('%s\n' % cmd)
                    else:
                        sh.write('%s\n' % cmd.replace('${SCRATCH_FOLDER}', ''))

    def get_job_name(self, name: str, chunk: str, soft=None):
        self.job_name = name + '.' + self.pjct
        if soft:
            self.job_name += '.' + ''.join([
                x for x in soft.dir.split('after_')[-1] if x not in 'aeiuoy'])
        self.job_name += '.' + chunk.replace('/', '_')

    def write_script(self, soft=None):
        if self.config.jobs:
            if soft:
                self.prep_script(soft.params)
            else:
                self.prep_script(self.config.params)
            self.call_cmd()
        else:
            self.job_fps.append('%s.sh' % splitext(self.sh)[0])

    def write_jobs(self, name: str, soft=None):
        for chunk, chunk_keys in self.chunks.items():
            self.get_sh(name, chunk, soft)
            self.write_chunks(chunk_keys, soft)
            self.get_job_name(name, chunk, soft)
            self.write_script(soft)

    def display(self):
        for db_soft, name_main in self.run.items():
            if len(self.run[db_soft]):
                print()
                soft_print = '========== %s scripts ========== ' % db_soft
                print(soft_print)
                self.scripts.append(soft_print)
            for name, main in name_main.items():
                main_print = '>%s\nsh %s' % (name, main)
                print(main_print)
                self.scripts.append(main_print)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import numpy as np
import pandas as pd


def read_metadata(
        meta_fp: str
) -> pd.DataFrame:
    """Read the metadata passed by the user.

    Parameters
    ----------
    meta_fp : str
        Metadata file path.

    Returns
    -------
    meta : pd.DataFrame
        Metadata table.
    """
    first_col = get_first_column(meta_fp)
    meta = pd.read_csv(meta_fp, dtype=str, sep='\t', low_memory=False)
    meta.rename(columns={first_col: 'sample_name'}, inplace=True)
    return meta


def get_first_column(
        meta_fp: str
) -> str:
    """Get the first column of the metadata file.

    Parameters
    ----------
    meta_fp : str
        Metadata file path.

    Returns
    -------
    first_column : str
        Name of the first column of the metadata input file.
    """
    line = ''
    with open(meta_fp) as f:
        for line in f:
            break
    if line.split():
        first_column = line.split()[0]
        return first_column
    else:
        raise IOError('File "%s" is empty' % meta_fp)


def get_cur_type(
        cur_input: str
) -> str:
    """Get the string "nucl" or "prot" depending
    on the type of data in the fasta file.

    Parameters
    ----------
    cur_input : str

    Returns
    -------
    typ : str
        Type of sequence data.
    """
    len_set = []
    with open(cur_input) as f:
        for ldx, line in enumerate(f):
            if line.startswith('>'):
                continue
            len_set.append(len(set(line.strip())))
            if ldx > 15:
                break
    if np.mean(len_set) < 6:
        typ = 'nucl'
    else:
        typ = 'prot'
    return typ




# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import yaml
import unittest
import pkg_resources
from os.path import isfile

import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal
import io
from unittest.mock import patch

from metagenomix.core.databases import ReferenceDatabases
from metagenomix.core.config import AnalysesConfig

RESOURCES = pkg_resources.resource_filename('metagenomix', 'resources')
FOLDER = pkg_resources.resource_filename('metagenomix', 'tests/unittests')


class TestDatabases(unittest.TestCase):

    def setUp(self) -> None:

        self.mar = '%s/databases/MAR' % FOLDER
        self.pfam = '%s/databases/PFAM' % FOLDER
        self.dbcan = '%s/databases/dbCAN' % FOLDER
        self.species = '%s/databases/dbCAN/species.txt' % FOLDER
        self.squeezemeta = '%s/databases/squeezemeta' % FOLDER
        self.wol = '%s/databases/wol' % FOLDER
        self.midas = '%s/databases/midas' % FOLDER
        self.metaphlan = '%s/databases/metaphlan' % FOLDER

        self.yaml = {'mar': {'path': self.mar},
                     'pfam': {'path': self.pfam, 'terms': ['oxin', 'anti']},
                     'dbcan': {'path': self.dbcan, 'taxa': {'A': self.species}},
                     'squeezemeta': {'path': self.squeezemeta},
                     'wol': {'path': self.wol},
                     'midas': {'path': self.midas},
                     'metaphlan': {'path': self.metaphlan}}
        self.databases_fp = '%s/yamls/databases.yml' % FOLDER
        with open(self.databases_fp, "w") as o:
            yaml.dump(self.yaml, o)
        self.config = {'databases_yml': self.databases_fp, 'midas_foci': {}}

        self.yaml_simple = {'A': {'path': 'path/to/A'},
                            'B': {'path': 'path/to/B'}}
        self.databases_simple_fp = '%s/yamls/databases_simple.yml' % FOLDER
        with open(self.databases_simple_fp, "w") as o:
            yaml.dump(self.yaml_simple, o)
        self.config_simple = {'databases_yml': self.databases_simple_fp}

        self.no_path = {'example1': {}, 'example2': {'something': 'here'}}
        self.no_path_fp = '%s/yamls/yaml_no_path.yml' % FOLDER
        with open(self.no_path_fp, "w") as o:
            yaml.dump(self.no_path, o)
        self.config_no_path = {'databases_yml': self.no_path_fp}

        self.no_config = {'databases_yml': None}

        self.paths = []
        self.configs = {}
        for database, database_d in self.yaml.items():
            self.cur_fp = '%s/yamls/databases_%s.yml' % (FOLDER, database)
            with open(self.cur_fp, "w") as o:
                yaml.dump({database: self.yaml[database]}, o)
            self.paths.append(self.cur_fp)
            self.configs[database] = {'databases_yml': self.cur_fp}

        self.meta_dir = '%s/dbCAN-seq' % self.dbcan
        self.cazyme_seq_list = '%s/CAZyme_seq_list' % self.meta_dir
        self.meta_fp = '%s/metadata.txt' % self.meta_dir
        self.meta = pd.DataFrame(
            [['GCF_000007365.1', 'Buchnera'], ['GCF_000021405.1', 'Borrelia'],
             ['GCF_001047655.1', 'Leptospira']], columns=['gen', 'genome_name'])
        self.meta.set_index('gen', inplace=True)

    @patch('sys.stdout', new_callable=io.StringIO)
    def assert_stdout(self, databases, expected, mock_stdout):
        databases.show_loaded_databases()
        self.assertEqual(mock_stdout.getvalue(), expected)

    def test_show_loaded_databases_prints(self):
        config = AnalysesConfig(**self.no_config)
        config.parse_yamls()
        databases = ReferenceDatabases(config)
        expected = 'No database passed using option `-d`\n'
        self.assert_stdout(databases, expected)

        config = AnalysesConfig(**self.config_no_path)
        config.parse_yamls()
        databases = ReferenceDatabases(config)
        expected = 'Databases in %s\n' % self.no_path_fp
        expected += '[Database] "example1" folder not found\n'
        expected += '[Database] "example2" folder not found\n'
        self.assert_stdout(databases, expected)

        config = AnalysesConfig(**self.config_simple)
        config.parse_yamls()
        databases = ReferenceDatabases(config)
        expected = 'Databases in %s\n' % self.databases_simple_fp
        expected += '- A: path/to/A\n'
        expected += '- B: path/to/B\n'
        self.assert_stdout(databases, expected)

        config = AnalysesConfig(**self.config)
        config.parse_yamls()
        databases = ReferenceDatabases(config)
        expected = 'Databases in %s\n' % self.databases_fp
        for k, v in sorted(self.yaml.items()):
            expected += '- %s: %s\n' % (k, v['path'])
        self.assert_stdout(databases, expected)

    def test_show_loaded_databases(self):

        config = AnalysesConfig(**self.no_config)
        config.parse_yamls()
        databases = ReferenceDatabases(config)
        res = databases.show_loaded_databases()
        self.assertFalse(res)

        config = AnalysesConfig(**self.config)
        config.parse_yamls()
        databases = ReferenceDatabases(config)
        res = databases.show_loaded_databases()
        self.assertTrue(res)

    def test_set_squeezemeta(self):
        config = AnalysesConfig(**self.config)
        config.parse_yamls()
        databases = ReferenceDatabases(config)
        databases.set_squeezemeta()
        self.assertEqual(databases.database, 'squeezemeta')

    def test_set_mar(self):
        pass

    def test_set_pfam(self):
        anti = 'PF12574.11__120_KDa_Rickettsia_surface_antigen'
        oxin = 'PF10417.12__C_terminal_domain_of_1_Cys_peroxiredoxin'
        hmms = {'oxin': {oxin: ['%s/oxin/%s.hmm' % (self.pfam, oxin),
                                '%s/oxin/%s.dmnd' % (self.pfam, oxin)]},
                'anti': {anti: ['%s/anti/%s.hmm' % (self.pfam, anti),
                                '%s/anti/%s.dmnd' % (self.pfam, anti)]}}
        hmms_pd = pd.DataFrame([
            ['1-cysPrx_C', 'PF10417.12', 'C-terminal domain of 1-Cys peroxiredoxin',
             '21.1; 21.1;', 'Domain', 40, np.nan, np.nan],
            ['120_Rick_ant', 'PF12574.11', '120 KDa Rickettsia surface antigen',
             '25; 25;', 'Family', 238, np.nan, np.nan]],
            columns=['ID', 'AC', 'DE', 'GA', 'TP', 'ML', 'CL', 'NE'])
        config = AnalysesConfig(**self.configs['pfam'])
        config.parse_yamls()
        databases = ReferenceDatabases(config)
        databases.set_pfam()
        self.assertEqual(databases.database, 'pfam')
        self.assertEqual(databases.hmms, hmms)
        assert_frame_equal(databases.hmms_pd, hmms_pd)

    def test_get_dbcan_hmms(self):
        expected = {}
        for hmm in ['a', 'b', 'c']:
            hmm_fp = '%s/%s.hmm' % (self.dbcan, hmm)
            with open(hmm_fp, 'w') as o: pass
            expected[hmm] = hmm_fp

        config = AnalysesConfig(**self.configs['dbcan'])
        config.parse_yamls()
        databases = ReferenceDatabases(config)
        databases.get_dbcan_hmms()
        self.assertEqual(databases.hmms, expected)

        for _, hmm_fp in expected.items():
            os.remove(hmm_fp)

        config = AnalysesConfig(**self.configs['dbcan'])
        config.parse_yamls()
        databases = ReferenceDatabases(config)
        databases.get_dbcan_hmms()
        self.assertEqual(databases.hmms, {})

    def test_write_dbcan_subset(self):
        config = AnalysesConfig(**self.configs['dbcan'])
        config.parse_yamls()
        databases = ReferenceDatabases(config)
        databases.dbcan_meta = self.meta

        folder = '%s/subset' % config.databases['dbcan']['path']
        os.makedirs(folder)
        os.makedirs(self.cazyme_seq_list)
        gcf_fas = '%s/GCF_000007365.1.fasta' % self.cazyme_seq_list
        with open(gcf_fas, 'w') as o:
            o.write('>header\nACGT\n')
        fas, dia = '%s/A.fa' % folder, '%s/A.dmnd' % folder
        expected = "diamond makedb --in %s -d %s\n" % (fas, dia)
        observed = databases.write_dbcan_subset(['Buchnera'], folder, 'A')
        self.assertEqual(observed, expected)
        self.assertTrue(isfile(fas))

        os.remove(fas)
        os.remove(gcf_fas)
        os.rmdir(self.cazyme_seq_list)
        os.rmdir(self.meta_dir)

        config = AnalysesConfig(**self.configs['dbcan'])
        config.parse_yamls()
        databases = ReferenceDatabases(config)
        with open(dia, 'w') as o: pass
        observed = databases.write_dbcan_subset(['Buchnera'], folder, 'A')
        self.assertEqual(observed, '')
        os.remove(dia)
        os.rmdir(folder)

    def test_set_dbcan_taxa(self):
        config = AnalysesConfig(**self.configs['dbcan'])
        config.parse_yamls()
        databases = ReferenceDatabases(config)
        databases.set_dbcan_taxa()
        self.assertEqual(databases.cazys, {})

        # folder = '%s/subset' % config.databases['dbcan']['path']
        # with open(self.species, 'w') as o:
        #     o.write('a_species\n')
        # databases.set_dbcan_taxa()
        # self.assertEqual(databases.cazys, {'A': folder})
        # os.remove(self.species)
        # os.rmdir(folder)

    def test_set_dbcan(self):
        config = AnalysesConfig(**self.configs['dbcan'])
        config.parse_yamls()
        databases = ReferenceDatabases(config)
        databases.set_dbcan()
        self.assertEqual(databases.database, 'dbcan')
        self.assertEqual(databases.hmms, {})
        assert_frame_equal(databases.dbcan_meta, pd.DataFrame())

        os.makedirs(self.meta_dir)
        self.meta.to_csv(self.meta_fp, sep='\t')
        config = AnalysesConfig(**self.configs['dbcan'])
        config.parse_yamls()
        databases = ReferenceDatabases(config)
        databases.set_dbcan()
        assert_frame_equal(databases.dbcan_meta, self.meta)
        os.remove(self.meta_fp)
        os.rmdir(self.meta_dir)

        expected = {}
        for hmm in ['a', 'b', 'c']:
            hmm_fp = '%s/%s.hmm' % (self.dbcan, hmm)
            with open(hmm_fp, 'w') as o: pass
            expected[hmm] = hmm_fp
        databases.set_dbcan()
        self.assertEqual(databases.hmms, expected)
        for _, hmm_fp in expected.items():
            os.remove(hmm_fp)

    def test_set_wol(self):
        pass

    def test_set_midas(self):
        pass

    def test_set_humann(self):
        pass

    def test_set_metaphlan(self):
        pass

    def tearDown(self) -> None:
        os.remove(self.databases_fp)
        os.remove(self.databases_simple_fp)
        os.remove(self.no_path_fp)


if __name__ == '__main__':
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import click

from metagenomix.monitoring import monitoring
from metagenomix import __version__


@click.command()
@click.option(
    "-i", "--i-fastq-illumina", multiple=True,
    help="Path to short Illumina reads fastq files folder(s).")
@click.option(
    "-j", "--i-fastq-pacbio", multiple=True,
    help="Path to long PacBio reads fastq files folder(s).")
@click.option(
    "-k", "--i-fastq-nanopore", multiple=True,
    help="Path to long MinION Nanopore reads fastq files folder.")
@click.option(
    "-m", "--m-metadata", required=True, help="Path to the metadata file.")
@click.option(
    "-o", "--o-out-dir", required=True, help="Path to pipeline output folder.")
@click.option(
    "-s", "--o-summary", required=True, help="Path to summary output file.")
@click.option(
    "-d", "--p-databases", show_default=True, help="Databases (yaml file).")
@click.option(
    "-p", "--p-pipeline", show_default=True, help="Tools to pipeline.")
@click.option(
    "-u", "--p-run-params", show_default=True, help="Server run parameters.")
@click.option(
    "-c", "--p-co-assembly", show_default=True,
    help="Metadata column(s) defining the co-assembly groups (yaml file).")
@click.option(
    "-t", "--p-strains", show_default=True,
    help="Species for strain-level analyses (yaml file).")
@click.option(
    "--show-params/--no-show-params", default=False, show_default=False,
    help="Show all possible parameters for all tools of your pipeline")
@click.option(
    "--force/--no-force", default=False, show_default=True,
    help="Check as if the re-writing of scripts for all commands was planned")
@click.option(
    "--verbose/--no-verbose", default=False, show_default=True,
    help="Whether to show input/outputs and other details")
@click.option(
    "--dev/--no-dev", default=False, show_default=True,
    help="For development...")
@click.version_option(__version__, prog_name="metagenomix")
def monitor(
        m_metadata,
        i_fastq_illumina,
        i_fastq_pacbio,
        i_fastq_nanopore,
        o_out_dir,
        o_summary,
        p_co_assembly,
        p_databases,
        p_pipeline,
        p_run_params,
        p_strains,
        show_params,
        force,
        verbose,
        dev
):
    """Check IO/job status of your pipeline configuration."""
    monitoring(
        meta_fp=m_metadata,
        illumina_dirs=i_fastq_illumina,
        pacbio_dirs=i_fastq_pacbio,
        nanopore_dirs=i_fastq_nanopore,
        output_dir=o_out_dir,
        summary_fp=o_summary,
        coassembly_yml=p_co_assembly,
        databases_yml=p_databases,
        pipeline_tsv=p_pipeline,
        user_params_yml=p_run_params,
        strains_yml=p_strains,
        show_params=show_params,
        force=force,
        verbose=verbose,
        dev=dev
    )



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
import re
import pkg_resources
import pandas as pd
import datetime as dt
from os.path import basename
from metagenomix.core.output import Output


def manager(**kwargs):
    """Run metagenomix to assist user renaming or discarding outputs.

    Parameters
    ----------
    kwargs : dict
        All arguments passed in command line, including defaults
    """
    print('\n === metagenomix manager ===\n')
    managing = Managed(**kwargs)

    print('* setting up the output file name')
    managing.get_out()
    print(' -> %s' % managing.out)

    print('* getting softwares to manage')
    managing.get_softs()

    print('* parsing folders to collect info about jobs and outputs')
    managing.parse_softs()

    for name, soft in managing.data.items():
        print()
        print('software:', name)
        print(pd.DataFrame(soft.jobs))


class Managed(object):

    def __init__(self, **kwargs) -> None:
        self.__dict__.update(kwargs)
        self.softs = {'res': {}, 'dir': set(), 'pip': set(), 'usr': set()}
        self.time = dt.datetime.now().strftime("%d/%m/%Y, %H") + 'h'
        self.manage_dir = '%s/_management' % self.folder
        self.roles = {}
        self.data = {}

    def _from_dir(self):
        self.softs['dir'].update([soft for soft in os.listdir(self.folder)])

    def _from_usr(self):
        for soft in self.softwares:
            if soft in self.softs['dir']:
                self.softs['usr'].add(soft)
            else:
                r = re.compile(soft)
                self.softs['usr'].update(
                    set(filter(r.match, list(self.softs['dir']))))

    def _from_res(self):
        res = pkg_resources.resource_filename("metagenomix", "resources")
        with open('%s/softwares.txt' % res) as f:
            for line in f:
                soft, role = line.strip().split('\t')
                self.roles.setdefault(role.split(' (')[0],set()).add(soft)
                self.softs['res'][soft] = role

    def _from_pipeline(self):
        if self.pipeline:
            with open(self.pipeline) as f:
                self.softs['pip'].update([line.strip().split()[-1] for line in
                                          f if line.strip() and line[0] != '#'])

    def _intersection(self) -> set:
        softs = self.softs['dir'] & set(self.softs['res'])
        if self.softs['pip'] | self.softs['usr']:
            softs = softs & (self.softs['pip'] | self.softs['usr'])
        return softs

    def _to_manage(self) -> dict:
        softs_intersection = self._intersection()
        role_softs = {}
        for role, softs in self.roles.items():
            cur_softs = softs & softs_intersection
            if cur_softs:
                role_softs[role] = {t: self.softs['res'][t] for t in cur_softs}
        return role_softs

    def show(self):
        print('Tools to manage per role:')
        for role, softs in self.softs.items():
            print('  [Role: %s]' % role)
            n = max([len(soft) for soft in softs])
            roles = set(softs.values())
            for soft, role_ in softs.items():
                print('\t- %s' % soft, end='')
                if role_ == role or len(roles) == 1:
                    print()
                else:
                    print(' ' * (n-len(soft)), '\t:', role_.split('(')[-1][:-1])

    def get_softs(self):
        self._from_dir()
        self._from_usr()
        self._from_res()
        self._from_pipeline()
        self.softs = self._to_manage()
        if self.verbose:
            self.show()

    def get_out(self):
        if self.out is None:
            path = self.time + '.txt'
        else:
            path = basename(self.out)
            if '/' in self.out:
                print('Using "%s" to write in "%s"' % (path, self.manage_dir))
        self.out = self.manage_dir + '/' + path

    def parse_softs(self):
        """An Output class instance is created for each software to manage,
        and placed as value to the dict with the software name of key."""
        for role, softs in self.softs.items():
            for soft in softs:
                self.data[soft] = Output(self.folder, soft)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import sys
from metagenomix._io_utils import to_do, status_update


def pool_cmd(
        self,
        tech: str,
        pool: str,
        paths: list,
        fasta: str,
        group: str
) -> None:
    """Write the pooling command and collect the output and io for FLASh.

    Parameters
    ----------
    self : Commands class instance
        .cmds
            Command lines
        .config
            Configurations
    tech: str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    pool : str
        Name of the pool
    paths : list
        Path the input fasta files
    fasta : str
        Path to an output fasta file
    group : str
        Name of the sample group within the pool
    """
    if self.config.force or to_do(fasta):
        cmd = ''
        for pdx, path in enumerate(paths):
            if pdx:
                cmd += 'cat %s >> %s\n' % (path, fasta)
            else:
                cmd += 'cat %s > %s\n' % (path, fasta)
        if cmd:
            self.cmds.setdefault((pool, group, tech), []).append(cmd)


def extension_paths(
        paths_to_merge: dict,
        fastqs: list
) -> None:
    """Fill the `paths_to_merge` dict with the fastq files to merge based on
    the matching of their extension which are the extensions to be expected
    based on the number of files available for the sample/technology and the
    previous steps of the pipeline.

    Parameters
    ----------
    paths_to_merge : dict
        Paths to the input files to merge per type of output
    fastqs : list
        Path the input files for the current sample and technology
    """
    extensions = {1: ['extendedFrags.fastq.gz', 'fastq.gz'],
                  2: ['notCombined_1.fastq.gz', 'notCombined_2.fastq.gz',
                      'R1.fastq.gz', 'R2.fastq.gz'],
                  3: ['extendedFrags.fastq.gz',
                      'notCombined_1.fastq.gz',
                      'notCombined_2.fastq.gz']}
    for fastq in fastqs:
        exts = extensions[len(fastqs)]
        for ext in exts:
            if fastq.endswith(ext):
                paths_to_merge.setdefault(ext, []).append(fastq)
                break
        else:
            sys.exit('[pooling] File "%s" has no extension %s' % (fastq, exts))


def collect_paths_to_merge(
        self,
        tech: str,
        sams: list,
) -> dict:
    """Get the list of input file paths to merge in order to constitute one
    file representing the current pool, and this for the current orientations,
    that can be 1 file (usually for long read data), 2 files (for paired-end
    illumina reads that have not been merged), or 3 files (for paired-end
    illimina reads that have been merged).

    Notes
    -----
    * "orientation" means all the per-sample files that are read 1 or
      read2, or in the case of merged reads, the "extended" or "notCombined"
      outputs.
    * the returned dict could be empty if there is no input for the
      current technology, in which case the pooling simply won't happen.

    Parameters
    ----------
    self : Commands class instance
        .inputs : dict
            Input files
    tech: str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    sams : list
        List of samples for the current pooling group

    Returns
    -------
    paths_to_merge : dict
        Paths to the input files to merge per type of output
    """
    paths_to_merge = {}
    for sam in sams:
        if (tech, sam) in self.inputs[sam] and self.inputs[sam][(tech, sam)]:
            fastqs = self.inputs[sam][(tech, sam)]
            extension_paths(paths_to_merge, fastqs)
    return paths_to_merge


def combine_single(
        paths_to_merge: dict
) -> None:
    """Combine the 'fastq.gz' file with the 'extendedFrags.fastq.gz' file
    if it happens that the samples' fastqs to combine are from samples that
    were both single and paired-end.

    Parameters
    ----------
    paths_to_merge : dict
        Paths to the input files to merge per type of output
    """
    if 'fastq.gz' in paths_to_merge and len(paths_to_merge) > 1:
        ext = 'extendedFrags.fastq.gz'
        paths_to_merge.setdefault(ext, []).extend(paths_to_merge['fastq.gz'])
        del paths_to_merge['fastq.gz']


def get_fasta_pools(
        self,
        tech: str,
        out: str,
        sams: list,
        pool: str,
        group: str
) -> list:
    """Collect the paths to the fasta file resulting from the pooling
    for the current group, including the preparation of the command for
    making this pool.

    Parameters
    ----------
    self : Commands class instance
        .cmds
            Command lines
        .soft.io : dict
            All files to I/O is scratch is used
        .config
            Configurations
    tech: str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    out : str
        Path the output folder
    sams : list
        Samples to pool
    pool : str
        Name of the pool
    group : str
        Name of the sample group within the pool

    Returns
    -------
    fasta_fps : list
        Paths to the fasta file resulting from the pooling for the current group
    """
    paths_to_merge = collect_paths_to_merge(self, tech, sams)
    if self.soft.params['pool_single_and_merged']:
        combine_single(paths_to_merge)
    fasta_fps = []
    for extension, paths in sorted(paths_to_merge.items()):
        status_update(self, tech, paths, group=group)
        fasta = pool_fasta(self, tech, out, extension, paths, pool, group)
        fasta_fps.append(fasta)
    return fasta_fps


def fasta_fp(
        out: str,
        group: str,
        extension: str
) -> str:
    """Get the path to the fasta file that results from merging.

    Parameters
    ----------
    out : str
        Path the output folder
    group : str
        Name of the sample group within the pool
    extension : str
        Extension to the files to merge

    Returns
    -------
    fasta : str
        Fasta file name for the merging result
    """
    fasta = '%s/%s.%s' % (out, group, extension)
    fasta = fasta.replace(' ', '_').replace('..', '.')
    return fasta


def add_to_pool_io(
        self,
        io: tuple,
        key: tuple,
        values: list
) -> None:
    """Add to the self.soft.io data structure the path to move to/from
    scratch for the pooling step specifically (which is pivotal between
    per sample processing to per pool/pool group).

    Parameters
    ----------
    self : Commands class instance
        .soft.io : dict
            All files to I/O is scratch is used
    io : tuple
        ('I', 'd'), ('I', 'f'), ('O', 'd'), or ('O', 'f')
    key : tuple
        (Name of the pool, Name of the pool group, Name pf the technology)
    values: list
        Paths to move to/from scratch for the pooling
    """
    if key not in self.soft.io:
        self.soft.io[key] = {('I', 'd'): set(), ('I', 'f'): set(),
                             ('O', 'd'): set(), ('O', 'f'): set()}
    self.soft.io[key][io].update(values)


def make_pool(
        self,
        tech: str,
        pool: str,
        group: str,
        paths: list,
        fasta: str
) -> None:
    """

    Parameters
    ----------
    self : Commands class instance
        .cmds
            Command lines
        .soft.io : dict
            All files to I/O is scratch is used
        .config
            Configurations
    tech: str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    pool : str
        Name of the pool
    group : str
        Name of the sample group within the pool
    paths : list
        Path to the input files to merge
    fasta : str
        Fasta file name for the merging result
    """
    # collect the inputs to pool as to move to scratch
    add_to_pool_io(self, ('I', 'f'), (pool, group, tech), paths)
    # collect the pooling command line
    pool_cmd(self, tech, pool, paths, fasta, group)


def pool_fasta(
        self,
        tech: str,
        out: str,
        extension: str,
        paths: list,
        pool: str,
        group: str
) -> str:
    """

    Parameters
    ----------
    self : Commands class instance
        .cmds
            Command lines
        .soft.io : dict
            All files to I/O is scratch is used
        .config
            Configurations
    tech: str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    out : str
        Path the output folder
    extension : str
        Extension to the files to merge
    paths : list
        Paths to the input files to merge
    pool : str
        Name of the pool
    group : str
        Name of the sample group within the pool

    Returns
    -------
    fasta : str
        Path to the file resulting from the pooling
    """
    # only pool if there is min 2 samples being merged
    if len(paths) > 1:
        fasta = fasta_fp(out, group, extension)
        make_pool(self, tech, pool, group, paths, fasta)
    else:
        fasta = paths[0]
    return fasta


def get_pools(
        self,
        pool: str,
        group: str,
        sams: list
) -> None:
    """Get the output fasta files returned by FLASh and write the commands.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for pools
        .inputs : dict
            Input files
        .cmds
            Command lines
        .soft.dirs : set
            All directories to create
        .soft.io : dict
            All files to I/O is scratch is used
        .config
            Configurations
    pool : str
        Name of the pool
    group : str
        Name of the sample group within the pool
    sams : list
        Samples to pool
    """
    # not possible to iterate over samples since pooling considers ALL samples
    for tech in self.config.techs:
        # initialize the data structure directly for the Commands instance
        out = '%s/%s/%s' % (self.dir, tech, pool)
        self.soft.dirs.add(out.replace('${SCRATCH_FOLDER}', ''))
        # get the fasta files of the pools (per orientation)
        fasta_fps = get_fasta_pools(self, tech, out, sams, pool, group)
        # fill the data structures
        add_to_pool_io(self, ('O', 'd'), (pool, group, tech), [out])
        add_to_pool_io(self, ('O', 'f'), (pool, group, tech), fasta_fps)
        self.soft.outputs[pool][group][tech] = fasta_fps


def pooling(
        self,
        pool: str
) -> None:
    """Create command lines for pooling samples' reads using FLASh.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for spades
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    pool : str
        Name of the pool
    """
    for group, group_pd in self.config.meta.groupby(pool):
        # get the full list of samples and the samples per pooling group
        sams = group_pd.index.tolist()
        self.pools[pool][group] = sams
        self.soft.outputs[pool][group] = {}
        # get the outputs for the current group and collect pooling commands
        get_pools(self, pool, group, sams)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import glob
import sys
import pkg_resources
from os.path import basename, isdir, splitext

from metagenomix._inputs import (sample_inputs, group_inputs,
                                 genome_key, genome_out_dir)
from metagenomix._io_utils import caller, io_update, to_do, status_update
from metagenomix.core.parameters import tech_params

RESOURCES = pkg_resources.resource_filename("metagenomix", "resources/scripts")


def prodigal_cmd(
        self,
        tech: str,
        fasta_fp: str,
        out: str
) -> str:
    """Create command lines for Prodigal.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for prodigal
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params : dict
            Parameters
        .config
            Configurations
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fasta_fp : str
        Path to the input fasta file
    out : str
        Path to the output folder

    Returns
    --------
    cmd : str
        Prodigal command
    """
    params = tech_params(self, tech)
    cmd = ''
    contig_fa = fasta_fp
    if fasta_fp.endswith('fastq.gz'):
        contig_fa = fasta_fp.replace('fastq.gz', 'fasta')
        cmd += 'seqtk seq -a %s > %s\n' % (fasta_fp, contig_fa)
    cmd = 'prodigal'
    cmd += ' -i %s' % contig_fa
    cmd += ' -a %s/protein.translations.fasta' % out
    cmd += ' -d %s/nucleotide.sequences.fasta' % out
    cmd += ' -s %s/potential.starts.fasta' % out
    cmd += ' -o %s/gene.coords.%s' % (out, params['f'])
    for param in ['f', 'p']:
        cmd += ' -%s %s' % (param, params[param])
    for boolean in ['c', 'm', 'n', 'q']:
        if params[boolean]:
            cmd += ' -%s' % params[boolean]
    return cmd


def get_prodigal(
        self,
        tech: str,
        fastas: dict,
        out_dir: str,
        sam_group: str,
) -> None:
    """Get the prodigal command and fill the pipeline data structures.

    Parameters
    ----------
    self : Commands class instance
        .outputs : dict
            All outputs
        .config
            Configurations
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastas : dict
        Paths to the input fasta file per sample name / data source
    out_dir : str
        Paths to the output folder
    sam_group : str
        Sample name or group for the current co-assembly
    """
    for genome, fasta in fastas.items():

        out = out_dir
        key = '_'.join([tech, sam_group])
        if genome:
            out += '/' + genome
            key += '_' + genome
        self.outputs['dirs'].append(out)
        self.outputs['outs'].setdefault((tech, sam_group), []).append(out)
        status_update(self, tech, [fasta[0]], group=sam_group, genome=genome)

        proteins = '%s/protein.translations.fasta' % out
        if self.config.force or to_do(proteins):
            cmd = prodigal_cmd(self, tech, fasta[0], out)
            self.outputs['cmds'].setdefault(key, []).append(cmd)
            io_update(self, i_f=fasta[0], o_d=out, key=key)
            self.soft.add_status(
                tech, self.sam_pool, 1, group=sam_group, genome=genome)
        else:
            self.soft.add_status(
                tech, self.sam_pool, 0, group=sam_group, genome=genome)


def prodigal(self) -> None:
    """Fast, reliable protein-coding gene prediction for prokaryotic genomes.

    References
    ----------
    Hyatt, Doug, et al. "Prodigal: prokaryotic gene recognition and
    translation initiation site identification." BMC bioinformatics 11.1 (
    2010): 1-11.

    Notes
    -----
    GitHub  : https://github.com/hyattpd/Prodigal
    Paper   : https://doi.org/10.1186/1471-2105-11-119

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for prodigal
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params : dict
            Parameters
        .config
            Configurations
    """
    if self.sam_pool in self.pools:
        for (tech, group), inputs in self.inputs[self.sam_pool].items():
            fastas = group_inputs(self, inputs)
            out_dir = '/'.join([self.dir, tech, self.sam_pool, group])
            get_prodigal(self, tech, fastas, out_dir, group)

    elif set(self.inputs) == {''}:
        for (tech, mags), inputs in self.inputs[''].items():
            fastas = group_inputs(self, inputs)
            out_dir = '/'.join([self.dir, tech, mags])
            get_prodigal(self, tech, fastas, out_dir, mags)

    else:
        tech_fastas = sample_inputs(self, ['pacbio', 'nanopore'])
        for tech, fastas in tech_fastas.items():
            out_dir = '/'.join([self.dir, tech, self.sam_pool])
            get_prodigal(self, tech, fastas, out_dir, self.sam_pool)


def macsyfinder_cmd(
        self,
        tech: str,
        proteins: str,
        sam_group: str,
        genome: str,
        out_dir: str,
        models_dir: str
) -> tuple:
    """Collect command for MacSyFinder.

    Parameters
    ----------
    self : Commands class instance
        .outputs : dict
            All outputs
        .soft.params : dict
            Parameters
        .config
            Configurations
    tech : str
        Technology: 'illumina', 'pacbio', 'nanopore', or hybrif naming
    proteins : str
        Path to the input proteins fasta file
    sam_group : str
        Sample name or group for the current co-assembly
    genome : str
        MAGs/Genomes folder name or empty string (for assembly contigs)
    out_dir : str
        Path to the output folder
    models_dir : str
        Path to the models folder

    Returns
    -------
    cmd : str
        MacSyFinder command
    outs : dict
        Paths to the output folder per model
    """
    cmd = ''
    proteins_fp = proteins
    params = tech_params(self, tech)
    if tech in ['illumina', 'pacbio', 'nanopore']:
        proteins_fp = '%s_edit.fasta' % proteins.replace('.fasta', '')
        cmd += '%s/header_space_replace.py -i %s -o %s --n\n' % (
            RESOURCES, proteins, proteins_fp)

    outs = {}
    for model in params['models']:
        model_dir = '%s/%s' % (out_dir, model)
        if not self.config.dev and not isdir(model_dir):
            self.soft.add_status(tech, self.sam_pool, 1, group=sam_group,
                                 message='no model %s' % model, genome=genome)
            continue
        outs[model] = model_dir
        self.outputs['dirs'].append(model_dir)
        res = '%s/macsyfinder.log' % model_dir
        if self.config.force or to_do(res):
            cmd += 'macsyfinder'

            if self.soft.prev == 'plass':
                cmd += ' --db-type unordered'
            elif '_spades_' in out_dir:
                cmd += ' --db-type ordered_replicon'
                cmd += ' --replicon-topology linear'
            elif '_drep_' in out_dir or '_metawrap' in out_dir:
                cmd += ' --db-type ordered_replicon'
                cmd += ' --replicon-topology circular'

            for param in [
                'e_value_search', 'i_evalue_sel', 'coverage_profile',
                'mandatory_weight', 'accessory_weight', 'exchangeable_weight',
                'redundancy_penalty', 'out_of_cluster'
            ]:
                cmd += ' --%s %s' % (param.replace('_', '-'), params[param])
            cmd += ' --out-dir %s' % model_dir
            cmd += ' --res-search-suffix _hmm.tsv'
            cmd += ' --res-extract-suffix _out.tsv'
            cmd += ' --worker %s' % params['cpus']
            cmd += ' --sequence-db %s' % proteins_fp
            cmd += ' --models-dir %s/models' % models_dir
            cmd += ' --models %s all' % model
            cmd += ' --verbosity\n'
            self.soft.add_status(
                tech, self.sam_pool, 1, group=sam_group, genome=genome)
        else:
            self.soft.add_status(
                tech, self.sam_pool, 0, group=sam_group, genome=genome)

    return cmd, outs


def get_macsyfinder(
        self,
        tech: str,
        fastas: dict,
        sam_group: str,
) -> None:
    """Get the MacSyFinder command and fill the pipeline data structures.

    Parameters
    ----------
    self : Commands class instance
        .databases : dict
            Databases
        .outputs : dict
            All outputs
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastas : dict
        Paths to the input fasta files per genome/MAG
    sam_group : str
        Sample name or group for the current co-assembly
    """
    for genome, fasta in fastas.items():

        key = genome_key(tech, sam_group, genome)
        out_dir = genome_out_dir(self, tech, fasta[0], sam_group, genome)
        self.outputs['dirs'].append(out_dir)

        proteins = fasta[0]
        if self.soft.prev == 'prodigal':
            proteins = '%s/protein.translations.fasta' % fasta[0]
        status_update(self, tech, [proteins], group=sam_group, genome=genome)

        models_dir = self.databases.paths['macsyfinder']
        cmd, outs = macsyfinder_cmd(
            self, tech, proteins, sam_group, genome, out_dir, models_dir)
        self.outputs['outs'].setdefault((tech, sam_group), []).append(outs)
        if cmd:
            self.outputs['cmds'].setdefault(key, []).append(cmd)
            io_update(self, i_f=proteins, i_d=models_dir, o_d=out_dir, key=key)


def macsyfinder(self) -> None:
    """MacSyFinder - Detection of macromolecular systems in protein datasets
    using systems modelling and similarity search.

    References
    ----------
    Abby, Sophie S., et al. "MacSyFinder: a program to mine genomes for
    molecular systems with an application to CRISPR-Cas systems." PloS one
    9.10 (2014): e110726.

    Notes
    -----
    GitHub  : https://github.com/gem-pasteur/macsyfinder
    Docs    : https://macsyfinder.readthedocs.io/en/latest/
    Paper   : https://doi.org/10.1371/journal.pone.0110726

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for MacSyFinder
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params : dict
            Parameters
        .config
            Configurations
    """
    if self.sam_pool in self.pools:
        if self.soft.prev != 'prodigal':
            sys.exit('[macsyfinder] Runs on protein data (prodigal only)')
        for (tech, group), inputs in self.inputs[self.sam_pool].items():
            fastas = group_inputs(self, inputs)
            get_macsyfinder(self, tech, fastas, group)

    elif set(self.inputs) == {''}:
        for (tech, mags), inputs in self.inputs[''].items():
            fastas = group_inputs(self, inputs)
            get_macsyfinder(self, tech, fastas, mags)

    else:
        prev = self.soft.prev
        if prev not in ['plass', 'prodigal']:
            sys.exit('[macsyfinder] Runs on protein data (plass or prodigal)')
        tech_fastas = sample_inputs(self)
        for tech, fastas in tech_fastas.items():
            get_macsyfinder(self, tech, fastas, self.sam_pool)


def write_hmms_cmd(self) -> tuple:
    """Write a file to contain on each line, the paths to the .hmm
    files to search as part of integronfinder.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to the output directory.
        .databases.hmms_dias : dict
            .hmm files per name of profile.

    Returns
    -------
    cmd = str
        Command to write the HMMs file
    hmms_fp : str
        Path to the file containing the paths to the .hmm files to search.
    """
    cmd = ''
    hmms_fp = ''
    if self.databases.hmms_dias:
        hmms_fp = '%s/hmm.txt' % self.dir
        first = True
        for i, (target, gen_hmm) in enumerate(self.databases.hmms_dias.items()):
            if self.config.dev and i == 1:
                break
            for j, (hmm, _) in enumerate(gen_hmm.values()):
                if self.config.dev and j == 1:
                    break
                if first:
                    cmd += 'echo -e "%s" > %s\n' % (hmm, hmms_fp)
                    first = False
                else:
                    cmd += 'echo -e "%s" >> %s\n' % (hmm, hmms_fp)
    return cmd, hmms_fp


def integronfinder_cmd(
        self,
        tech: str,
        fastx: str,
        out: str,
) -> str:
    """Collect command for integron_finder.

    Parameters
    ----------
    self : Commands class instance
        .databases : dict
            Path to the reference databases
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastx : str
        Path to the input fasta/fastq(.gz) file
    out : str
        Path to the output folder

    Returns
    -------
    cmd : str
        integron_finder command
    """
    params = tech_params(self, tech)

    cmd = ''
    fasta = fastx
    if fastx.endswith('fastq.gz') or fastx.endswith('fastq'):
        fasta = fastx.replace('fastq.gz', 'fasta')
        cmd += 'seqtk seq -a %s > %s\n' % (fastx, fasta)

    fasta_out = '%s_min%snt.fasta' % (splitext(fasta)[0], params['min_length'])
    hmms_cmd, hmms = write_hmms_cmd(self)
    if hmms_cmd:
        cmd += hmms_cmd

    cmd += '\n%s/filter_on_length.py' % RESOURCES
    cmd += ' -i %s' % fasta
    cmd += ' -o %s' % fasta_out
    cmd += ' -t %s\n' % params['min_length']

    cmd += 'integron_finder'
    for boolean in [
        'local_max', 'promoter_attI', 'mute', 'pdf', 'gbk', 'union_integrases'
    ]:
        if params[boolean]:
            cmd += ' --%s' % boolean.replace('_', '-')
    cmd += ' --verbose'
    cmd += ' --outdir %s' % out
    cmd += ' --cpu %s' % params['cpus']
    if self.databases.hmms_dias:
        cmd += ' --func-annot'
        cmd += ' --path-func-annot %s' % hmms
    if params['prot_file']:
        cmd += ' --prot-file %s' % params['prot_file']
    if params['attc_model']:
        cmd += ' --attc-model %s' % params['attc_model']
        cmd += ' --evalue-attc %s' % params['evalue_attc']
        cmd += ' --max-attc-size %s' % params['max_attc_size']
        cmd += ' --min-attc-size %s' % params['min_attc_size']
    if params['topology_file']:
        cmd += ' --topology-file %s' % params['topology_file']
    else:
        if self.soft.prev == 'plass':
            cmd += ' --topology linear'
        elif '_spades_' in out:
            cmd += ' --topology linear'
        elif '_drep_' in out or '_metawrap' in out:
            cmd += ' --topology circ'
    cmd += ' %s' % fasta_out
    return cmd


def get_integronfinder(
        self,
        tech: str,
        fastas: dict,
        sam_group: str,
) -> None:
    """Get the integron_finder command and fill the pipeline data structures.

    Parameters
    ----------
    self : Commands class instance
        .databases : dict
            Databases
        .outputs : dict
            All outputs
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastas : dict
        Paths to the input fasta files per genome/MAG
    sam_group : str
        Sample name or group for the current co-assembly
    """
    for genome, fasta_ in fastas.items():

        key = genome_key(tech, sam_group, genome)
        out = genome_out_dir(self, tech, fasta_[0], sam_group, genome)
        self.outputs['dirs'].append(out)
        self.outputs['outs'].setdefault((tech, sam_group), []).append(out)

        fasta = fasta_[0]
        if self.soft.prev == 'prodigal':
            fasta = '%s/nucleotide.sequences.fasta' % fasta_[0]
        status_update(self, tech, [fasta], group=sam_group, genome=genome)

        fpo = '%s/Results_Integron_Finder_mysequences/mysequences.summary' % out
        if self.config.force or to_do(fpo):
            cmd = integronfinder_cmd(self, tech, fasta, out)
            self.outputs['cmds'].setdefault(key, []).append(cmd)
            io_update(self, i_f=fasta, o_d=out, key=key)
            self.soft.add_status(
                tech, self.sam_pool, 1, group=sam_group, genome=genome)
        else:
            self.soft.add_status(
                tech, self.sam_pool, 0, group=sam_group, genome=genome)


def integronfinder(self) -> None:
    """Finds integrons in DNA sequences.

    Integrons are major genetic element, notorious for their major
    implication in the spread of antibiotic resistance genes. More generally,
    integrons are gene-capturing platform, whose broader evolutionary role
    remains poorly understood. IntegronFinder is able to detect with high
    accuracy integron in DNA sequences. It is accurate because it combines
    the use of HMM profiles for the detection of the essential protein,
    the site-specific integron integrase, and the use of Covariance Models
    for the detection of the recombination site, the attC site.

    References
    ----------
    Cury, Jean, et al. "Identification and analysis of integrons and cassette
    arrays in bacterial genomes." Nucleic acids research 44.10 (2016):
    4539-4550.

    Notes
    -----
    GitHub  : https://github.com/gem-pasteur/Integron_Finder
    Docs    : https://integronfinder.readthedocs.io/en/latest/
    Paper   : https://doi.org/10.1093/nar/gkw319

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for integron_finder
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params : dict
            Parameters
        .config
            Configurations
    """
    if self.sam_pool in self.pools:
        for (tech, group), inputs in self.inputs[self.sam_pool].items():
            fastas = group_inputs(self, inputs)
            get_integronfinder(self, tech, fastas, group)

    elif set(self.inputs) == {''}:
        for (tech, mags), inputs in self.inputs[''].items():
            fastas = group_inputs(self, inputs)
            get_integronfinder(self, tech, fastas, mags)

    else:
        tech_fastas = sample_inputs(self)
        for tech, fastas in tech_fastas.items():
            get_integronfinder(self, tech, fastas, self.sam_pool)


# def custom_cmd(
#         self,
#         input_file,
#         out_dir,
#         dia_hmm
# ) -> list:
#     outs = {}
#     for target, gene_hmm_dia in self.databases.hmms_dias.items():
#         if target == 'cazy':
#             continue
#         for gene, (hmm, dia) in gene_hmm_dia.items():
#             sam_dir = '%s/%s' % (out_dir, gene)
#             self.outputs['dirs'].append(sam_dir)
#             out = '%s/%s.tsv' % (sam_dir, self.sam_pool)
#             if to_do(out):
#                 if dia_hmm == 'hmm':
#                     stdout = '%s_hmmer.out' % splitext(out)[0]
#                     io_update(self, i_f=hmm, o_f=[out, stdout], key=target)
#                     cmd = 'hmmsearch'
#                     cmd += ' --cut_tc'
#                     cmd += ' --tblout %s' % out
#                     cmd += ' -o %s' % stdout
#                     cmd += ' --cpu %s' % self.soft.params['cpus']
#                     cmd += ' %s %s' % (hmm, input_file)
#                 else:
#                     io_update(self, i_f=dia, o_f=out, key=target)
#                     tmp = '%s/%s_tmp' % (sam_dir, self.sam_pool)
#                     cmd = 'mkdir -p %s\n' % tmp
#                     cmd += 'diamond blastp'
#                     cmd += ' -d %s' % dia
#                     cmd += ' -q %s' % input_file
#                     cmd += ' -o %s' % out
#                     cmd += ' -k 1 --id 80'
#                     cmd += ' -p %s' % self.soft.params['cpus']
#                     cmd += ' -t %s' % tmp
#                 self.outputs['cmds'].setdefault(target, []).append(cmd)
#             else:
#                 io_update(self, i_f=out, key=target)
#             out2 = out.replace('.tsv', '_contigs.tsv')
#             if to_do(out2):
#                 io_update(self, o_f=out2, key=target)
#                 cmd = 'extract_custom_searched_contigs.py'
#                 cmd += ' -i %s' % out
#                 cmd += ' -o %s' % out2
#                 self.outputs['cmds'].setdefault(target, []).append(cmd)
#             outs.setdefault(target, []).extend([out, out2])
#     return outs


def search_cmd(
        self,
        tech: str,
        sam_group: str,
        proteins: str,
        out_dir: str,
        key: str
) -> dict:
    """Dispatch the command line creation for hmmer or diamond and
    for the different reference databases.

    Parameters
    ----------
    self : Commands class instance
        .sam : str
            Sample or co-assembly pool name
        .outputs : dict
            All outputs
        .soft.params : dict
            Parameters
        .config
            Configurations
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    sam_group : str
        Sample name or group for the current co-assembly
    proteins : str
        Path to the input proteins fasta file
    out_dir : str
        Path to the main output folder
    key : str
        Concatenation of variables names for the current analytic level

    Returns
    -------
    outs : dict
        Paths to the output folder per database
    """
    outs = {}
    params = tech_params(self, tech)
    tmp = '$TMPDIR/%s_%s_%s' % (self.soft.name, self.sam_pool, key)
    module_call = caller(self, __name__)
    for db, db_paths in params['databases'].items():
        for db_path in db_paths:
            db_base = splitext(basename(db_path))[0]
            db_out_dir = '%s/%s' % (out_dir, db_base)
            self.outputs['dirs'].append(db_out_dir)

            out = '%s/%s.tsv' % (db_out_dir, db_base)
            outs.setdefault((db, db_base), []).append(db_out_dir)

            if self.config.force or to_do(out):
                cmd = module_call(params, proteins, db_path, out, tmp)
                io_update(self, i_f=proteins, o_d=db_out_dir, key=key)
                self.outputs['cmds'].setdefault(key, []).append(cmd)
                self.soft.add_status(
                    tech, self.sam_pool, 1, group=sam_group, genome=db)
            else:
                self.soft.add_status(
                    tech, self.sam_pool, 0, group=sam_group, genome=db)

    return outs


def diamond(
        params: dict,
        fp: str,
        db_path: str,
        out: str,
        tmp_dir: str
) -> str:
    """Collect the command line for DIAMOND.

    Parameters
    ----------
    params : dict
        Parameters for diamond and the current technology
    fp : str
        Path to the input fasta file
    db_path : str
        Path to the reference hmm database
    out : str
        Path to the output file
    tmp_dir : str
        Path to the temporary folder

    Returns
    -------
    cmd : str
        diamond command
    """
    cmd = 'mkdir -p %s\n' % tmp_dir
    cmd += 'diamond blastp'
    cmd += ' --db %s' % db_path
    cmd += ' --query %s' % fp
    cmd += ' --out %s' % out
    if params['mode']:
        cmd += ' --%s' % params['mode']
    if params['top']:
        cmd += ' --top %s' % params['top']
    else:
        cmd += ' -k %s' % params['max_target_seqs']
    cmd += ' --threads %s' % params['cpus']
    for param in ['id', 'strand', 'masking', 'evalue',
                  'query_cover', 'subject_cover']:
        cmd += ' --%s %s' % (param.replace('_', '-'), params[param])
    cmd += ' --tmpdir %s\n' % tmp_dir
    return cmd


def hmmer(
        params: dict,
        fp: str,
        db_path: str,
        out: str,
        _: str
) -> str:
    """Collect the command line for HMMER.

    Parameters
    ----------
    params : dict
        Parameters for hmmer and the current technology
    fp : str
        Path to the input fasta file
    db_path : str
        Path to the reference hmm database
    out : str
        Path to the output file
    _ : str
        Unused argument

    Returns
    -------
    cmd : str
        hmmsearch command
    """
    stdout = '%s_per-sequence.out' % splitext(out)[0]
    domout = '%s_per-domain.out' % splitext(out)[0]
    cmd = 'hmmsearch'
    cmd += ' -o %s' % stdout
    cmd += ' --tblout %s' % out
    cmd += ' --domtblout %s' % domout
    for boolean in ['nobias', 'noali', 'max']:
        if params[boolean]:
            cmd += ' --%s' % boolean
    for cut in ['cut_ga', 'cut_nc', 'cut_tc']:
        if params[cut]:
            cmd += ' --%s' % cut
            break
    for param in ['E', 'Z', 'domE', 'domZ', 'textw']:
        cmd += ' --%s %s' % (param, params[param])
    cmd += ' --cpu %s' % params['cpus']
    cmd += ' %s' % db_path
    cmd += ' %s' % fp
    return cmd


def get_search(
        self,
        tech: str,
        fastas: dict,
        sam_group: str
) -> None:
    """Get the DIAMOND or HMMER search command
    and fill the pipeline data structures.

    Parameters
    ----------
    self : Commands class instance
        .databases : dict
            Databases
        .outputs : dict
            All outputs
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastas : dict
        Paths to the input fasta file per sample name / data source
    sam_group : str
        Sample name or group for the current co-assembly
    """
    for genome, fasta in fastas.items():

        key = genome_key(tech, sam_group, genome)
        out_dir = genome_out_dir(self, tech, fasta[0], sam_group, genome)
        self.outputs['dirs'].append(out_dir)

        proteins = fasta[0]
        if self.soft.prev == 'prodigal':
            proteins = '%s/protein.translations.fasta' % fasta[0]
        status_update(self, tech, [proteins], group=sam_group, genome=genome)

        outs = search_cmd(self, tech, sam_group, proteins, out_dir, key)
        if outs:
            self.outputs['outs'].setdefault((tech, sam_group), {}).update(outs)


def search(self) -> None:
    """Create command lines for custom searches based on DIAMOND or HMMER.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for searches (DIAMOND or HMMER)
        .pool : str
            Co-assembly pool name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params : dict
            Parameters
        .soft.prev : str
            Previous software
        .config
            Configurations
    """
    # This function splits the name of the software and calls as function (
    # from this module) the part of the software name that follows the first
    # underscore, e.g. software "search_diamond" would call `diamond()`
    name = self.soft.name
    if self.sam_pool in self.pools:
        if self.soft.prev != 'prodigal':
            sys.exit('[%s] Runs on protein data (plass, prodigal...)' % name)
        for (tech, group), inputs in self.inputs[self.sam_pool].items():
            fastas = group_inputs(self, inputs)
            get_search(self, tech, fastas, group)

    elif set(self.inputs) == {''}:
        if self.soft.prev != 'prodigal':
            sys.exit('[%s] Runs on protein data (plass, prodigal...)' % name)
        for (tech, mags), inputs in self.inputs[''].items():
            fastas = group_inputs(self, inputs)
            get_search(self, tech, fastas, mags)

    else:
        prev = self.soft.prev
        if prev not in ['plass', 'prodigal']:
            sys.exit('[%s] Runs on protein data (plass, prodigal...)' % name)
        tech_fastas = sample_inputs(self)
        for tech, fastas in tech_fastas.items():
            get_search(self, tech, fastas, self.sam_pool)


def prokka_cmd(
        self,
        contigs: str,
        out_dir: str,
        pref: str,
        config: dict,
        cols: list
) -> str:
    """Collect the command line for Prokka.

    Parameters
    ----------
    self : Commands class instance
        .soft.params : dict
            Parameters
    contigs : str
        Path to the contigs input file
    out_dir : str
        Path to the output folder
    pref : str
        Prefix to add to the output files
    config : dict
        One field#:taxon dict obtained by parsing the user-defined config
    cols : list
        Columns of the taxonomic config: 'genus', 'species', 'strain', 'plasmid'

    Returns
    -------
    cmd : str
        Prokka command
    """
    cmd = 'prokka'
    cmd += ' --mincontiglen %s' % self.soft.params['mincontiglen']
    cmd += ' --cpus %s' % self.soft.params['cpus']
    for boolean in ['metagenome', 'notrna', 'norrna']:
        if self.soft.params[boolean]:
            cmd += ' --%s' % boolean.replace('_', '-')
    cmd += ' --force'
    for col in cols:
        if col and config[col]:
            cmd += ' --%s %s' % (col, config[col])
    cmd += ' --prefix %s' % pref
    cmd += ' --outdir %s' % out_dir
    cmd += ' %s' % contigs
    return cmd


def prokka_config(
        self,
        cols: list
) -> list:
    """Parse the user-defined taxonomic config file to collect the taxonomic
    levels to annotate more precisely using Prokka.

    Parameters
    ----------
    self : Commands class instance
        .soft.params : dict
            Parameters
    cols : list
        Columns of the taxonomic config: 'genus', 'species', 'strain', 'plasmid'

    Returns
    -------
    configs : list
        List of field#:taxon dicts obtained by parsing the user-defined config
    """
    configs = []
    cols += ['proteins']
    if 'taxa' in self.soft.params:
        with open(self.soft.params['taxa']) as f:
            for ldx, line in enumerate(f):
                ls = line.strip('\n').split('\t')
                if not ldx:
                    continue
                configs.append({cols[idx]: val for idx, val in enumerate(ls)})
    return configs


def prokka_configs(self) -> tuple:
    """Get the config files for the different run of Prokka at
    different taxonomic levels.

    Parameters
    ----------
    self : Commands class instance
        .soft.params : dict
            Parameters

    Returns
    -------
    configs : list
        Dicts of config files per taxonomic level
    cols : list
        Taxonomic levels
    """
    if self.soft.params['config']:
        cols = ['genus', 'species', 'strain', 'plasmid']
        configs = prokka_config(self, cols)
    else:
        cols = ['']
        configs = [{'': 'no_config'}]
    return configs, cols


def get_prokka(
        self,
        tech: str,
        group: str,
        inputs: list,
        out_dir: str
) -> str:
    """Get the Prokka commands for the different configs
    (at each taxonomic level).

    Parameters
    ----------
    self : Commands class instance
        .soft.params : dict
            Parameters
        .config
            Configurations
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    group : str
        Name of the current co-assembly
    inputs : list
        Paths to the input files
    out_dir : str
        Path to the output folder

    Returns
    -------
    cmd : str
        Prokka command
    """
    cmd = ''
    configs, cols = prokka_configs(self)
    for config in configs:

        status_update(self, tech, [inputs[0]], group=group, genome=config)
        pref = '_'.join([config[x] for x in cols if config[x]])
        if config.get('proteins'):
            pref += '_%s' % splitext(basename(config['proteins']))[0]
        file_out = '%s/%s.out' % (out_dir, pref)
        if self.config.force or to_do(file_out):
            cmd += prokka_cmd(self, inputs[0], out_dir, pref, config, cols)
            self.soft.add_status(
                tech, self.sam_pool, 1, group=group, genome=config)
        else:
            self.soft.add_status(
                tech, self.sam_pool, 0, group=group, genome=config)

    return cmd


def prokka(self) -> None:
    """Prokka: rapid prokaryotic genome annotation.
    Whole genome annotation is the process of identifying features of
    interest in a set of genomic DNA sequences, and labelling them with
    useful information. Prokka is a software tool to annotate bacterial,
    archaeal and viral genomes quickly and produce standards-compliant
    output files,

    References
    ----------
    Seemann, Torsten. "Prokka: rapid prokaryotic genome annotation."
    Bioinformatics 30.14 (2014): 2068-2069.

    Notes
    -----
    GitHub  : https://github.com/tseemann/prokka
    Paper   : https://doi.org/10.1093/bioinformatics/btu153

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for prokka
        .pool : str
            Co-assembly pool name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params : dict
            Parameters
        .config
            Configurations
    """
    if self.sam_pool in self.pools:
        for (tech, group), inputs in self.inputs[self.sam_pool].items():
            out_dir = '/'.join([self.dir, tech, self.sam_pool, group])
            self.outputs['dirs'].append(out_dir)
            self.outputs['outs'].setdefault(group, []).append(out_dir)

            cmd = get_prokka(self, tech, group, inputs, out_dir)
            if cmd:
                tech_group = '_'.join([tech, group])
                self.outputs['cmds'].setdefault(tech_group, []).append(cmd)
                io_update(self, i_f=inputs[0], o_d=out_dir, key=tech_group)


def barrnap_cmd(
        self,
        tech: str,
        fastx: str,
        out: str
) -> str:
    """Collect the command line for Barrnap.

    Parameters
    ----------
    self : Commands class instance
        .soft.params : dict
            Parameters
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastx : str
        Path to the input fasta file
    out : str
        Path to the output file

    Returns
    -------
    cmd : str
        Barrnap command
    """
    params = tech_params(self, tech)
    cmd = ''
    fasta = fastx
    if fastx.endswith('fastq.gz') or fasta.endswith('fastq'):
        fasta = fasta.replace('fastq.gz', 'fasta').replace('fastq', 'fasta')
        cmd += 'seqtk seq -a %s > %s\n' % (fastx, fasta)
    cmd += 'barrnap'
    cmd += ' --kingdom %s' % params['kingdom']
    cmd += ' --threads %s' % params['cpus']
    cmd += ' --reject %s' % params['reject']
    cmd += ' --lencutoff %s' % params['lencutoff']
    cmd += ' --evalue %s' % params['evalue']
    if params['incseq']:
        cmd += ' --incseq'
    cmd += ' --outseq %s' % out
    cmd += ' %s' % fasta
    return cmd


def ccfind_cmd(
        self,
        tech: str,
        fastx: str,
        out_dir: str
) -> str:
    """Collect the command line for ccfind.

    Parameters
    ----------
    self : Commands class instance
        .soft.params : dict
            Parameters
        .config
            Configurations
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastx : str
        Path to the input fasta file
    out_dir : str
        Path to the output folder

    Returns
    -------
    cmd : str
        ccfind command
    """
    params = tech_params(self, tech)
    cmd = ''
    if self.config.force:
        cmd += 'if [ -d "%s" ]; then rm -rf %s; fi\n' % (out_dir, out_dir)
    fasta = fastx
    if fastx.endswith('fastq.gz') or fasta.endswith('fastq'):
        fasta = fasta.replace('fastq.gz', 'fasta').replace('fastq', 'fasta')
        cmd += 'seqtk seq -a %s > %s\n' % (fastx, fasta)
    cmd += 'ccfind %s %s' % (fasta, out_dir)
    cmd += ' --terminal-fragment-size %s' % params['terminal_fragment_size']
    cmd += ' --min-percent-identity %s' % params['min_percent_identity']
    cmd += ' --min-aligned-length %s' % params['min_aligned_length']
    cmd += ' --ncpus %s' % params['cpus']
    if params['preserve_tmpdir']:
        cmd += ' --preserve-tmpdir'
    return cmd


def get_ccfind(
        self,
        tech: str,
        fastas: dict,
        sam_group: str
) -> None:
    """Get the ccfind command and fill the pipeline data structures.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for ccfind
        .pool : str
            Co-assembly pool name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params : dict
            Parameters
        .config
            Configurations
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastas : dict
        Path(s) to the input fasta file(s) per genome/MAGs
    sam_group : str
        Sample name or name of a co-assembly pool's group
    """
    for genome, fasta in fastas.items():

        out_dir = genome_out_dir(self, tech, fasta[0], sam_group, genome)
        self.outputs['dirs'].append(out_dir)
        self.outputs['outs'].setdefault((tech, sam_group), []).append(out_dir)
        status_update(self, tech, [fasta[0]], group=sam_group, genome=genome)

        out = '%s/circ.detected.list' % out_dir
        if self.config.force or not to_do(out):
            cmd = ccfind_cmd(self, tech, fasta[0], out_dir)
            key = genome_key(tech, sam_group, genome)
            self.outputs['cmds'].setdefault(key, []).append(cmd)
            io_update(self, i_f=fasta[0], o_d=out_dir, key=key)
            self.soft.add_status(
                tech, self.sam_pool, 1, group=sam_group, genome=genome)
        else:
            self.soft.add_status(
                tech, self.sam_pool, 0, group=sam_group, genome=genome)


def get_barrnap(
        self,
        tech: str,
        fastas: dict,
        sam_group: str
) -> None:
    """Get the Barrnap command and fill the pipeline data structures.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for Barrnap
        .pool : str
            Co-assembly pool name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params : dict
            Parameters
        .config
            Configurations
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    fastas : dict
        Path(s) to the input fasta file(s) per genome/MAGs
    sam_group : str
        Sample name or name of a co-assembly pool's group
    """
    for genome, fasta in fastas.items():

        key = genome_key(tech, sam_group, genome)

        out_dir = genome_out_dir(self, tech, fasta[0], sam_group, genome)
        self.outputs['dirs'].append(out_dir)

        out = '%s/rrna_seqs.fas' % out_dir
        self.outputs['outs'].setdefault((tech, sam_group), []).append(out)
        status_update(self, tech, [fasta[0]], group=sam_group, genome=genome)

        if self.config.force or not to_do(out):
            cmd = barrnap_cmd(self, tech, fasta[0], out)
            self.outputs['cmds'].setdefault(key, []).append(cmd)
            io_update(self, i_f=fasta[0], o_d=out_dir, key=key)
            self.soft.add_status(
                tech, self.sam_pool, 1, group=sam_group, genome=genome)
        else:
            self.soft.add_status(
                tech, self.sam_pool, 0, group=sam_group, genome=genome)


def dispatch(self) -> None:
    """Run either barrnap or ccfind on the per-sample fastq files,
    or on the fasta resulting from the assembly, on per co-assembly MAGs for
    the genome dereplication step.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for filtering
        .pool : str
            Co-assembly pool name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params : dict
            Parameters
        .config
            Configurations
    """
    __plasmid_tool__ = getattr(sys.modules[__name__], 'get_%s' % self.soft.name)

    if self.sam_pool in self.pools:
        for (tech, group), inputs in self.inputs[self.sam_pool].items():
            fastas = group_inputs(self, inputs)
            __plasmid_tool__(self, tech, fastas, group)
    else:
        tech_fastas = sample_inputs(self, ['pacbio', 'nanopore'])
        for tech, fastas in tech_fastas.items():
            __plasmid_tool__(self, tech, fastas, self.sam_pool)


def barrnap(self) -> None:
    """Dispatch the collection of barrnap commands using generic function.

    Barrnap predicts the location of ribosomal RNA genes in genomes. It
    supports bacteria (5S,23S,16S), archaea (5S,5.8S,23S,16S), metazoan
    mitochondria (12S,16S) and eukaryotes (5S,5.8S,28S,18S).
    It takes FASTA DNA sequence as input, and write GFF3 as output. It uses
    the new nhmmer tool that comes with HMMER 3.1 for HMM searching in
    RNA:DNA style. Multithreading is supported and one can expect roughly
    linear speed-ups with more CPUs.

    References
    ----------
    Seemann T
    barrnap 0.9 : rapid ribosomal RNA prediction
    https://github.com/tseemann/barrnap

    Notes
    -----
    GitHub  : https://github.com/tseemann/barrnap

    Parameters
    ----------
    self : Commands class instance
    """
    dispatch(self)


def ccfind(self) -> None:
    """Dispatch the collection of ccfind commands using generic function.

    ccfind - Circular Complete genome FINDer.
    ccfind is a general tool to detect circular complete genomes with clues
    of terminal redundancy, originally designed for identification
    of complete virus genomes from metagenome assembly. It can be used for
    any contig/genome, but cutoff values should be carefully considered.
    It should be noted that terminal redundancy (circularity) does not
    necessarily mean completion of the sequence. Partial genomes might be
    detected as circular contigs for some reasons (e.g., sequence repeats).

    References
    ----------
    Nishimura, Yosuke, et al. "Environmental viral genomes shed new light on
    virus-host interactions in the ocean." MSphere 2.2 (2017): e00359-16.

    Notes
    -----
    GitHub  : https://github.com/yosuken/ccfind
    Paper   : https://doi.org/10.1128/mSphere.00359-16

    Parameters
    ----------
    self : Commands class instance
    """
    dispatch(self)


def antismash_cmd(
        self,
        fasta: str,
        out_dir: str
) -> str:
    """Collect the command line for Antismash.

    Parameters
    ----------
    self : Commands class instance
        .soft.params : dict
            Parameters
        .soft.prev : str
            Previous software
    fasta : str
        Path to the input genome fasta file
    out_dir : str
        Path to the output folder

    Returns
    -------
    cmd : str
        Antismash command
    """
    cmd = '\nantismash'
    for boolean in ['rre', 'asf', 'cassis', 'pfam2go', 'tigrfam', 'fullhmmer',
                    'cc_mibig', 'cb_general', 'smcog_trees', 'clusterhmmer',
                    'cb_subclusters', 'cb_knownclusters']:
        if self.soft.params[boolean]:
            cmd += ' --%s' % boolean.replace('_', '-')
    cmd += ' --tta-threshold %s' % self.soft.params['tta_threshold']
    cmd += ' --genefinding-tool %s' % self.soft.params['genefinding_tool']
    cmd += ' --html-description after_%s' % self.soft.prev
    cmd += ' --taxon %s' % self.soft.params['taxon']
    cmd += ' --cpus %s' % self.soft.params['cpus']
    cmd += ' --output-basename %s' % splitext(basename(fasta))[0]
    cmd += ' --html-title "%s"' % splitext(basename(fasta))[0]
    cmd += ' --output-dir %s' % out_dir
    cmd += ' %s\n' % fasta
    return cmd


def get_antismash(
        self,
        fastas: dict,
        tech: str,
        sam_group: str
) -> None:
    """

    Parameters
    ----------
    self : Commands class instance
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    fastas : dict
        Paths to the input fasta files per genome/MAG
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    sam_group : str
        Sample name or name of a co-assembly pool's group
    """
    for genome, dirs in fastas.items():

        fasta = dirs[0]
        out_dir = genome_out_dir(self, tech, fasta, sam_group, genome)
        self.outputs['dirs'].append(out_dir)
        self.outputs['outs'].setdefault((tech, sam_group), []).append(out_dir)

        key = genome_key(tech, sam_group, genome)
        if genome:
            condition = to_do(folder=fasta)
        else:
            condition = to_do(fasta)
        if condition:
            status_update(self, tech, [fasta], group=sam_group, genome=genome)

        if self.config.force or not glob.glob('%s/*' % out_dir):
            cmd = antismash_cmd(self, fasta, out_dir)
            self.outputs['cmds'].setdefault(key, []).append(cmd)
            io_update(self, i_f=fasta, o_d=out_dir, key=key)
            self.soft.add_status(
                tech, self.sam_pool, 1, group=sam_group, genome=genome)
        else:
            self.soft.add_status(
                tech, self.sam_pool, 0, group=sam_group, genome=genome)


def antismash(self) -> None:
    """Antibiotics and Secondary Metabolite Analysis SHell.
    antiSMASH allows the rapid genome-wide identification, annotation and
    analysis of secondary metabolite biosynthesis gene clusters in bacterial
    and fungal genomes. It integrates and cross-links with a large number of
    in silico secondary metabolite analysis tools that have been published
    earlier.

    References
    ----------
    Blin, Kai, et al. "antiSMASH 6.0: improving cluster detection and
    comparison capabilities." Nucleic Acids Research 49.W1 (2021): W29-W35.

    Notes
    -----
    GitHub  : https://github.com/antismash
    Paper   : https://doi.org/10.1093/nar/gkab335

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for Antismash
        .pool : str
            Co-assembly pool name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params : dict
            Parameters
        .soft.prev : str
            Previous software
        .config
            Configurations
    """
    if self.sam_pool in self.pools:
        for (tech, group), inputs in self.inputs[self.sam_pool].items():
            fastas = group_inputs(self, inputs)
            get_antismash(self, fastas, tech, group)

    elif self.soft.prev == 'drep':
        for (tech, bin_algo), inputs in self.inputs[''].items():
            fastas = group_inputs(self, inputs)
            get_antismash(self, fastas, tech, bin_algo)
    else:
        tech_fastas = sample_inputs(self, ['pacbio', 'nanopore'])
        for tech, fastas in tech_fastas.items():
            get_antismash(self, fastas, tech, self.sam_pool)


def tiara(self) -> None:
    """Deep-learning-based approach for identification of eukaryotic
    sequences in the metagenomic data powered by PyTorch. The sequences are
    classified in two stages:
    - In the first stage, the sequences are classified to classes: archaea,
      bacteria, prokarya, eukarya, organelle and unknown.
    - In the second stage, the sequences labeled as organelle in the first
      stage are classified to either mitochondria, plastid or unknown.

    Notes
    -----
    GitHub  : https://github.com/ibe-uw/tiara
    Paper   : https://doi.org/10.1093/bioinformatics/btab672

    Parameters
    ----------
    self : Commands class instance
        .soft.prev : str
            Previous software in the pipeline
        .dir : str
            Path to pipeline output folder for tiara
        .sam_pool : str
            Co-assembly name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    if self.soft.prev != 'spades':
        sys.exit('[tiara] can only be run on assembly output')
    for (tech, group), spades in self.inputs[self.sam_pool].items():
        out_dir = '/'.join([self.dir, tech, self.sam_pool, group])
        self.outputs['dirs'].append(out_dir)
        self.outputs['outs'][group] = out_dir
        out_fp = '%s/classifications.txt' % out_dir
        status_update(self, tech, [spades[0]], group=group)

        if self.config.force or to_do(out_fp):
            cmd = 'cd %s\n' % out_dir
            cmd += 'tiara'
            cmd += ' --input %s' % spades[0]
            cmd += ' --output %s' % out_fp
            cmd += ' --threads %s' % self.soft.params['cpus']
            for boolean in ['probabilities', 'verbose', 'gzip']:
                if self.soft.params[boolean]:
                    cmd += ' --%s' % boolean
            for param in ['min_len', 'first_stage_kmer', 'second_stage_kmer']:
                cmd += ' --%s %s' % (param, self.soft.params[param])
            for param in ['prob_cutoff', 'to_fasta']:
                cmd += ' --%s %s' % (param, ' '.join(self.soft.params[param]))
            self.outputs['cmds'].setdefault(group, []).append(cmd)
            io_update(self, i_f=spades[0], o_d=out_dir, key=group)
            self.soft.add_status(tech, self.sam_pool, 1, group=group)
        else:
            self.soft.add_status(tech, self.sam_pool, 0, group=group)

# def write_dbcan_subset(
#         self,
#         taxa: list,
#         folder: str,
#         name: str
# ) -> None:
#     """Write the fasta file suset to the target features and
#     the command to make it a diamond database.
#
#     Parameters
#     ----------
#     self : Commands class instance
#         .databases : dict
#             Path to the reference databases
#         .outputs : dict
#             All outputs
#     taxa : list
#         Taxa from the user file.
#     folder : str
#         dbCAN-Seq folder.
#     name : str
#         Current subset name.
#
#     Returns
#     -------
#     cmd : str
#         Command to make the diamond db from the subsets fasta file.
#     """
#     path = self.databases['dbcan']
#     cmd = ""
#     fas_fp = '%s/%s.fa' % (folder, name)
#     dia_fp = '%s.dmnd' % splitext(fas_fp)[0]
#     io_update(self, i_f=fas_fp, o_f=dia_fp)
#     if to_do(dia_fp):
#         with open(fas_fp, 'w') as o:
#             taxon_found = False
#             for taxon in taxa:
#                 meta_taxon_pd = self.databases.dbcan_meta.loc[
#                     self.databases.dbcan_meta.genome_name.str.contains(taxon),:]
#                 if not meta_taxon_pd.shape[0]:
#                     continue
#                 gcf_dir = '%s/dbCAN-seq/CAZyme_seq_list' % path
#                 if to_do(folder=gcf_dir):
#                     os.makedirs(gcf_dir)
#                 for gcf in set(meta_taxon_pd.index):
#                     gcf_fas = '%s/%s.fasta' % (gcf_dir, gcf)
#                     if not to_do(gcf_fas):
#                         for e in read(gcf_fas, 'fasta'):
#                             o.write('>%s\n%s\n' % (e.metadata['id'], e))
#                         taxon_found = True
#         if taxon_found:
#             cmd = "diamond makedb --in %s -d %s\n" % (fas_fp, dia_fp)
#     self.outputs['cmds'].append(cmd)
#
#
# def set_dbcan_taxa(self) -> None:
#     """
#
#     Parameters
#     ----------
#     self : Commands class instance
#         .soft.params : dict
#             Parameters
#     """
#     for name, fp in self.soft.params['taxa'].items():
#         taxa = list(reads_lines(fp))
#         if not taxa:
#             continue
#         folder = '%s/subsets' % self.dir
#         if to_do(folder=folder):
#             os.makedirs(folder)
#         self.databases.cazys[name] = folder
#         write_dbcan_subset(self, taxa, folder, name)
#
#
# def cazy(self) -> None:
#     """
#
#     Parameters
#     ----------
#     self : Commands class instance
#     """
#     set_dbcan_taxa(self)
#
#
# def prepare_ioncom_inputs(
#         self,
#         ion_com: str,
#         fp: str
# ) -> str:
#     """Create the preformatting command lines before running IonCom.
#
#     Parameters
#     ----------
#     self : Commands class instance
#         .dir : str
#             Path to pipeline output folder for IonCom
#         .sam : str
#             Sample name
#         .inputs : dict
#             Input files
#         .outputs : dict
#             All outputs
#         .soft.params : dict
#             Parameters
#         .config
#             Configurations
#     tech : str
#         Technology: 'illumina', 'pacbio', or 'nanopore'
#     ion_com : str
#         Path to the IonCom standalone
#     fp : str
#         Path to the output file
#     """
#     cmd = 'prepare_ioncom_inputs.py'
#     cmd += ' -d %s -p %s -n 1' % (ion_com, fp)
#     cmd += '%s/run_IonCom.pl' % ion_com
#     return cmd
#
#
# def ioncom(self) -> None:
#     """Create command lines for IonCom.
#
#     Parameters
#     ----------
#     self : Commands class instance
#         .dir : str
#             Path to pipeline output folder for IonCom
#         .sam : str
#             Sample name
#         .inputs : dict
#             Input files
#         .outputs : dict
#             All outputs
#         .soft.params : dict
#             Parameters
#         .config
#             Configurations
#     """
#     # -------------------------------------------------------
#     # --- WILL NEED TO BE REVISED AFTER I-TASSER DOWNLOAD ---
#     # -------------------------------------------------------
#     i_tasser_libs = self.databases.ioncom['itasser']
#     i_tasser = self.soft.params['itasser']
#     ion_com = '%s/IonCom_standalone' % self.databases.paths['ioncom']
#
#     if self.sam_pool in self.pools:
#         for group in self.pools[self.sam_pool]:
#             o_dir, fp = get_out_dir(self, self.sam_pool, group)
#             cmd = prepare_ioncom_inputs(self, ion_com, fp)
#             self.outputs['cmds'].setdefault(group, []).append(cmd)
#             output = '%s/output' % ion_com
#             cmd = 'for i in %s/*rehtml\n' % output
#             cmd += 'do\n'
#             cmd += '    mv %s/*rehtml %s/.\n' % (output, o_dir)
#             cmd += 'done\n'
#             cmd += 'for i in %s/*/*/prob_*.txt\n' % output
#             cmd += 'do\n'
#             cmd += '    mkdir -p "$i"\n'
#             cmd += 'done'
#             self.outputs['cmds'].setdefault(group, []).append(cmd)
#             self.outputs['dir'].append(o_dir)
#             self.outputs['outs'].setdefault(self.sam_pool, []).append(o_dir)
#             io_update(self, i_f=fp, i_d=[ion_com, i_tasser, i_tasser_libs],
#                       o_d=o_dir, key=group)
#
#



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import os
from os.path import basename, splitext

from metagenomix._io_utils import io_update


def get_meta_groupings(self) -> dict:
    """

    Returns
    -------
    meta_groups : dict
        Samples per group's unique value per group variable
    """
    meta_groups = {}
    meta = self.config.meta.copy()
    for group in self.soft.params['groups']:  # for each group from user params
        meta_groups[group] = {}
        group_type = str(meta[group].dtype)
        # can not use group/variables that are numerical
        if group_type != 'object':
            continue
        for factor in set(meta[group]):
            sams = meta.loc[meta[group] == factor, :].index.tolist()
            meta_groups[group][factor] = sams
        sample_folder = '%s/%s' % (self.dir, group)
        self.outputs['cmds'].append('mkdir -p %s' % sample_folder)
    return meta_groups


def get_input_files(self):
    input_files = {}
    all_input_files = []
    # fdx = 0

    #
    # GET THE RIGHT OUTPUT
    #
    for root, dirs, files in os.walk('%s/%s' % (self.dir, self.soft.prev)):
        for fil in files:
            if 'bin' in fil and 'metawrap_' in root.split('/')[-1] and '_bins' in root.split('/')[-1]:
                # fdx += 1
                # if fdx > 5:
                #     continue
                path = root + '/' + fil
                sam = root.split(self.soft.prev)[-1].split('/')[3]
                input_files.setdefault(sam, []).append(path)
                all_input_files.append(path)
                io_update(self, i_f=all_input_files)
    return input_files


def metamarker(self) -> None:
    """Prepare the command and collect outputs, io and dirs for Metamarker.

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder
        .outputs : dict
            All outputs
        .soft
            Software class instances and attributes
        .config
            Configuration class instance
    """
    # note that metamarker can run on genomes data after e.g.:
    #   - metawrap_ref
    #   - other
    #  which should be contained in some user_params

    meta_groups = get_meta_groupings(self)
    input_files = get_input_files(self)
    data = metamarker_step1(self, meta_groups, input_files)
    metamarker_step2(self, data)
    metamarker_step3(self, data, meta_groups)
    metamarker_step4(self, data)
    metamarker_step5(self, data, meta_groups)
    metamarker_step6(self, data)


def metamarker_step1(self, meta_groups: dict, input_files: dict) -> dict:

    data = {'npzs': [], 'nsamples': [], 'sample_folders': []}
    for meta_group, samples in meta_groups.items():
        self.outputs['cmds'].append('\n')
        sample_folder = '%s/%s' % (self.dir, meta_group)
        n_samples = 0
        meta_group_fas = []
        for sdx, sam in enumerate(samples):
            if sam not in input_files:
                continue
            for idx, input_file in enumerate(input_files[sam]):
                cur_fas = '%s/%s.%s.fasta' % (
                    sample_folder, sam, splitext(basename(input_file))[0])
                cur_cmd = '/bin/cp %s %s' % (input_file, cur_fas)
                n_samples += 1
                self.outputs['cmds'].append(cur_cmd)
                meta_group_fas.append(cur_fas)

        out_npz = '%s/outStep1/%s/%s.npz' % (self.dir, meta_group, meta_group)
        out_folder = '%s/outStep1/%s' % (self.dir, meta_group)
        cur_cmd = 'Step1_MakeHashTable.py'
        cur_cmd += ' --sample %s' % sample_folder
        cur_cmd += ' --output %s' % out_folder
        cur_cmd += ' --npz %s' % out_npz
        mkdir_cmd = 'mkdir -p %s' % out_folder

        cmd = '\n\necho "%s"\n' % mkdir_cmd
        cmd += mkdir_cmd
        cmd += '\necho "%s"\n' % cur_cmd
        cmd += cur_cmd
        self.outputs['cmds'].append(cmd)

        data['npzs'].append(out_npz)
        data['nsamples'].append(n_samples)
        data['sample_folders'].append(sample_folder)

    return data


def metamarker_step2(self, data: dict):
    out_folder = '%s/out_Step2' % self.dir
    cur_cmd = 'Step2_MakeShortMarker.py'
    cur_cmd += ' --caseHash %s' % data['npzs'][0]
    cur_cmd += ' --controlHash %s' % data['npzs'][1]
    cur_cmd += ' --spade /home/flejzerowicz/softs/SPAdes-3.13.0-Linux/bin'
    cur_cmd += ' --output %s' % out_folder
    cur_cmd += ' --threads %s' % self.soft.params['cpus']
    mkdir_cmd = 'mkdir -p %s' % out_folder
    cmd = '\n\necho "%s"\n' % mkdir_cmd
    cmd += mkdir_cmd
    cmd += '\necho "%s"\n' % cur_cmd
    cmd += cur_cmd
    self.outputs['cmds'].append(cmd)
    data['step2_outs'] = ['%s/Short_Markers_case.fasta' % out_folder,
                          '%s/Short_Markers_control.fasta' % out_folder]


def metamarker_step3(self, data: dict, meta_groups: dict):
    data['step3_outs'] = []
    self.outputs['cmds'].append('\n')
    for mdx, meta_group in enumerate(meta_groups.keys()):
        cur_out = '%s/out_Step3/%s' % (self.dir, meta_group)
        data['step3_outs'].append(cur_out)
        tmpdir = '$TMPDIR/MetaMarker/%s' % meta_group
        cur_cmd = 'Step3_MakeLongMarker.py'
        cur_cmd += ' --short_marker %s' % data['step2_outs'][mdx]
        cur_cmd += ' --sample %s' % data['sample_folders'][mdx]
        cur_cmd += ' --tmp %s' % tmpdir
        cur_cmd += ' --output %s' % cur_out
        cur_cmd += ' --spade /home/flejzerowicz/softs/SPAdes-3.13.0-Linux/bin'
        cur_cmd += ' --usearch /home/flejzerowicz/softs'
        cur_cmd += ' --gt /home/flejzerowicz/usr/local/genometools/bin'
        cur_cmd += ' --threads %s' % self.soft.params['cpus']
        cur_cmd += '  --id %s' % self.soft.params['identity']
        mkdir_cmd = 'mkdir -p %s' % cur_out
        cmd = '\n\necho "%s"\n' % mkdir_cmd
        cmd += mkdir_cmd
        cmd += '\necho "%s"\n' % cur_cmd
        cmd += cur_cmd
        self.outputs['cmds'].append(cmd)


def metamarker_step4(self, data: dict):
    cur_out = '%s/out_Step4' % self.dir
    cur_cmd = 'Step4_MarkerCleaning.py'
    cur_cmd += ' --caseMarker %s' % data['step3_outs'][0]
    cur_cmd += ' --caseSize %s' % data['nsamples'][0]
    cur_cmd += ' --controlMarker %s' % data['step3_outs'][1]
    cur_cmd += ' --controlSize %s' % data['nsamples'][1]
    cur_cmd += ' --output %s' % cur_out
    cur_cmd += ' --cdhit /home/flejzerowicz/softs/cd-hit-v4.6.8-2017-1208'
    mkdir_cmd = 'mkdir -p %s' % cur_out
    cmd = '\n\necho "%s"\n' % mkdir_cmd
    cmd += mkdir_cmd
    cmd += '\necho "%s"\n' % cur_cmd
    cmd += cur_cmd
    self.outputs['cmds'].append(cmd)


def metamarker_step5(self, data: dict, meta_groups: dict):
    data['step5_outs'] = []
    self.outputs['cmds'].append('\n')
    for mdx, meta_group in enumerate(meta_groups.keys()):
        cur_out = '%s/out_Step5/%s' % (self.dir, meta_group)
        tmpdir = '$TMPDIR/MetaMarker/Step5/%s' % meta_group
        data['step5_outs'].append(cur_out)
        cur_cmd = 'Step5_MarkerAbundace.py'
        cur_cmd += ' --marker %s/out_Step4/Selected_Markers.fasta' % self.dir
        cur_cmd += ' --sample %s' % data['sample_folders'][mdx]
        cur_cmd += ' --output %s' % cur_out
        cur_cmd += ' --tmp %s' % tmpdir
        cur_cmd += ' --alnlength 50'
        cur_cmd += ' --id %s' % self.soft.params['identity']
        cur_cmd += ' --bowtie2 /home/flejzerowicz/usr/miniconda3/bin'
        cur_cmd += ' --threads %s' % self.soft.params['cpus']
        cur_cmd += ' --removeTemp'
        mkdir_cmd = 'mkdir -p %s' % cur_out
        cmd = '\n\necho "%s"\n' % mkdir_cmd
        cmd += mkdir_cmd
        cmd += '\necho "%s"\n' % cur_cmd
        cmd += cur_cmd
        self.outputs['cmds'].append(cmd)


def metamarker_step6(self, data: dict):
    cur_out = '%s/out_Step6' % self.dir
    cur_cmd = 'Step6_MarkerRank.py'
    cur_cmd += ' --case %s' % data['step5_outs'][0]
    cur_cmd += ' --control %s' % data['step5_outs'][1]
    cur_cmd += ' --marker %s/out_Step4/Selected_Markers.fasta' % self.dir
    cur_cmd += ' --cdhit /home/flejzerowicz/softs/cd-hit-v4.6.8-2017-1208'
    cur_cmd += ' --output %s' % cur_out
    mkdir_cmd = 'mkdir -p %s' % cur_out
    cmd = '\n\necho "%s"\n' % mkdir_cmd
    cmd += mkdir_cmd
    cmd += '\necho "%s"\n' % cur_cmd
    cmd += cur_cmd
    self.outputs['cmds'].append(cmd)
    step6_outs = ['%s/%s' % (cur_out, x) for x in [
        'Final_Marker_Case.csv', 'Final_Marker_Case.fasta',
        'Final_Marker_Control.csv', 'Final_Marker_Control.fasta',
        'Marker_Stats.csv']]
    io_update(self, o_f=step6_outs)
    self.outputs['outs'].extend(step6_outs)



# ----------------------------------------------------------------------------
# Copyright (c) 2022, Franck Lejzerowicz.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import sys
from metagenomix._io_utils import io_update, to_do, status_update
from metagenomix.core.parameters import tech_params


def quast_data(
        self,
        tech: str,
        pool: str,
        group_samples: dict
) -> tuple:
    """Collect the paths to the contigs for all the co-assembly groups of the
    current co-assembly pool name, and the respective metadata values
    qualifying these groups for the user-defined metadata columns.

    Parameters
    ----------
    self : Commands class instance
        .inputs : dict
            Input files
        .soft.params
            Parameters
        .config
            Configurations
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    pool : str
        Name of the co-assembly pool
    group_samples : dict
        Name of a group for the current co-assembly pool: list of pooled samples

    Returns
    -------
    contigs : list
        Paths to the contigs assembly files
    labels : list
        Labels for the co-assembly groups of the current pool
    """
    contigs, labels = [], []
    for group, assembly_outputs in sorted(self.inputs[pool].items()):
        if group[0] != tech:
            continue
        contigs.append(assembly_outputs[0])
        if self.soft.params['label']:
            samples = group_samples[group[1]]
            vals = self.config.meta.loc[samples, self.soft.params['label']]
            labels.append('__'.join(sorted(set(vals))))
    return contigs, labels


def quast_cmd(
        self,
        tech: str,
        pool: str,
        group_samples: dict,
        out_dir: str
) -> tuple:
    """Collect the QUAST command.

    Parameters
    ----------
    self : Commands class instance
        .inputs : dict
            Input files
        .soft.params
            Parameters
        .config
            Configurations
    tech : str
        Technology: 'illumina', 'pacbio', or 'nanopore'
    pool : str
        Name of the co-assembly pool
    group_samples : dict
        Name of a group for the current co-assembly pool: list of pooled samples
    out_dir : str
        Path to the output folder

    Returns
    -------
    cmd : str
        QUAST command
    contigs : list
        Paths to the contigs assembly files
    """
    contigs, labels = quast_data(self, tech, pool, group_samples)
    status_update(self, tech, contigs, pool=pool)

    cmd = 'metaquast.py'
    cmd += ' %s' % ' '.join(contigs)
    if labels:
        cmd += ' --labels %s' % ','.join(labels)
    cmd += ' --output-dir %s' % out_dir
    cmd += ' --min-contig %s' % self.soft.params['min_contig']
    cmd += ' --ambiguity-usage %s' % self.soft.params['ambiguity_usage']
    for boolean in ['circos', 'glimmer', 'rna_finding',
                    'conserved_genes_finding', 'space_efficient']:
        if self.soft.params[boolean]:
            cmd += ' --%s' % boolean.replace('_', '-')
    if self.soft.params['fast']:
        cmd += ' --fast'
    else:
        for boolean in ['no_check', 'no_plots', 'no_html', 'no_icarus',
                        'no_snps', 'no_gc', 'no_sv', 'no_read_stats']:
            if self.soft.params[boolean]:
                cmd += ' --%s' % boolean.replace('_', '-')
    return cmd, contigs


def quast(self) -> None:
    """QUAST stands for QUality ASsessment Tool. It evaluates
    genome/metagenome assemblies by computing various metrics. The current
    QUAST toolkit includes the general QUAST tool for genome assemblies,
    MetaQUAST, the extension for metagenomic datasets, QUAST-LG,
    the extension for large genomes (e.g., mammalians), and Icarus,
    the interactive visualizer for these tools.

    References
    ----------
    Gurevich, Alexey, et al. "QUAST: quality assessment tool for genome
    assemblies." Bioinformatics 29.8 (2013): 1072-1075.

    Notes
    -----
    GitHub  : https://github.com/ablab/quast
    Docs    : http://quast.sourceforge.net/
    Paper   : https://doi.org/10.1093/bioinformatics/btt086

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for QUAST
        .pools : dict
            Pool name : { pool group : [pooled samples] }
        .pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    for pool, group_sams in self.pools.items():

        for tech in [x[0] for x in self.inputs[pool]]:
            tech_pool = '%s_%s' % (tech, pool)
            out_dir = '%s/%s/%s' % (self.dir, tech, pool)
            if self.soft.params['label']:
                out_dir += '_%s' % self.soft.params['label']
            self.outputs['dirs'].append(out_dir)
            self.outputs['outs'][pool] = out_dir

            if self.config.force or not to_do('%s/report.html' % out_dir):
                cmd, contigs = quast_cmd(self, tech, pool, group_sams, out_dir)
                self.outputs['cmds'].setdefault(tech_pool, []).append(cmd)
                io_update(self, i_f=contigs, o_d=out_dir, key=tech_pool)
                self.soft.add_status(tech, pool, 1)
            else:
                self.soft.add_status(tech, pool, 0)


def spades_cmd(
        self,
        techs_inputs: dict,
        techs: list,
        tmp: str,
        out: str,
        log: str,
        hybrid: str,
        group: str
) -> tuple:
    """Create command lines for SPAdes.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    techs_inputs : dict
        Dict of the technology:paths to the input fasta/fastq.gz files
    techs : list
        Technologies to use for the (hybrid) assembly
    tmp : str
        Temporary folder for SPAdes
    out : str
        Path to the output folder
    log : str
        Path to the output log file
    hybrid : str
        Name of the hybrid assembly
    group : str
        Name of a group for the current co-assembly pool

    Returns
    -------
    cmd : str
        SPAdes.py command
    inputs : list
        Path to the input files
    """
    cmd = 'mkdir -p %s\n' % tmp
    cmd += 'spades.py'
    cmd += ' -o %s' % out
    cmd += ' -k %s' % ','.join(map(str, self.soft.params['k']))
    cmd += ' --memory %s' % self.soft.params['mem']
    cmd += ' --threads %s' % self.soft.params['cpus']
    cmd += ' --tmp-dir %s' % tmp
    for boolean in ['meta', 'careful', 'disable_gzip_output',
                    'disable_rr', 'cov_cutoff']:
        if self.soft.params[boolean]:
            cmd += ' --%s' % boolean.replace('_', '-')

    cmd += ' --checkpoints %s' % self.soft.params['checkpoints']

    if self.soft.params['only_assembler']:
        cmd += ' --only-assembler'
    elif self.soft.params['only_error_correction']:
        cmd += ' --only-error-correction'

    mode = self.soft.name.split('_')[-1]
    if mode == 'bio':
        cmd += ' --bio'
    elif mode == 'plasmid':
        cmd += ' --plasmid'
    elif mode == 'metaviral':
        cmd += ' --metaviral'

    inputs = []
    if self.soft.params['continue'] and not to_do(log):
        cmd += ' --continue'
    else:
        for tech, fastqs in techs_inputs.items():
            if tech not in techs:
                continue
            status_update(self, hybrid, fastqs, group=group)
            inputs.extend(fastqs)
            if tech != 'illumina':
                if len(fastqs):
                    cmd += ' --%s %s' % (tech, ' '.join(fastqs))
            else:
                for fastq in fastqs:
                    if len(fastq) == 1:
                        cmd += ' -s %s' % fastq
                    else:
                        if 'extendedFrags' in fastq:
                            cmd += ' --merge %s' % fastq
                        elif 'notCombined_1' in fastq:
                            cmd += ' -1 %s' % fastq
                        elif 'notCombined_2' in fastq:
                            cmd += ' -2 %s' % fastq
                        elif 'R1.fastq' in fastq:
                            cmd += ' -2 %s' % fastq
                        elif 'R2.fastq' in fastq:
                            cmd += ' -2 %s' % fastq
    cmd += '\nrm -rf %s\n' % tmp
    return cmd, inputs


def hybridize_tech(
        self,
        techs_inputs: dict
) -> tuple:
    """Get the technologies to use for hybrid assembly and the folder name
    for this hybrid assembly.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    techs_inputs : dict
        Dict of the technology:paths to the input fasta/fastq.gz files

    Returns
    -------
    techs : list
        Technologies to use for the hybrid assembly
    hybrid : str
        Name of the hybrid assembly
    """
    techs = []
    for tech in self.soft.params['hybrid']:
        if tech in techs_inputs:
            techs.append(tech)

    hybrid = '_'.join(techs)
    if len(techs) > 1:
        hybrid = 'hybrid__%s' % hybrid
    return techs, hybrid


def spades(self) -> None:
    """SPAdes - St. Petersburg genome assembler - is an assembly toolkit
    containing various assembly pipelines.

    References
    ----------
    Prjibelski, Andrey, et al. "Using SPAdes de novo assembler." Current
    protocols in bioinformatics 70.1 (2020): e102.
    Bankevich, Anton, et al. "SPAdes: a new genome assembly algorithm and its
    applications to single-cell sequencing." Journal of computational biology
    19.5 (2012): 455-477.

    Notes
    -----
    GitHub  : https://github.com/ablab/spades
    Docs    : https://cab.spbu.ru/software/spades/
    Paper   : https://doi.org/10.1002/cpbi.102

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for SPAdes
        .pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    # iterate over inputs
    for group, techs_inputs in self.inputs[self.sam_pool].items():

        techs, hybrid = hybridize_tech(self, techs_inputs)

        # make output folder path
        tmp = '$TMPDIR/spades_%s_%s_%s' % (hybrid, group, self.sam_pool)
        out = '%s/%s/%s/%s' % (self.dir, hybrid, self.sam_pool, group)
        # collect as folder to create under 'dirs'
        self.outputs['dirs'].append(out)

        # get expected output files
        before_rr = '%s/before_rr.fasta' % out
        contigs = '%s/contigs.fasta' % out
        first_pe = '%s/first_pe_contigs.fasta' % out
        scaffolds = '%s/scaffolds.fasta' % out
        log = '%s/spades.log' % out
        outs = [contigs, before_rr, first_pe, scaffolds, log]
        self.outputs['outs'][(hybrid, group)] = outs

        # check if there is a need to run
        if self.config.force or to_do(contigs):
            cmd, inputs = spades_cmd(
                self, techs_inputs, techs, tmp, out, log, hybrid, group)
            self.outputs['cmds'].setdefault(group, []).append(cmd)
            io_update(self, i_f=inputs, i_d=out, o_d=out, key=group)
            self.soft.add_status(hybrid, self.sam_pool, 1, group=group)
        else:
            self.soft.add_status(hybrid, self.sam_pool, 0, group=group)


def megahit_cmd(
        self,
        inputs: list,
        group: str,
        tmp: str,
        out: str,
) -> str:
    """Create command lines for SPAdes.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    inputs : list
        Paths to the input fasta/fastq.gz files
    group : str
        Name of a group for the current co-assembly pool
    tmp : str
        Temporary folder for SPAdes
    out : str
        Path to the output folder

    Returns
    -------
    cmd : str
        SPAdes.py command
    inputs : list
        Path to the input files
    """
    cmd = 'mkdir -p %s\n' % tmp
    cmd += 'megahit'
    cmd += ' --tmp-dir %s' % tmp
    cmd += ' --out-prefix %s' % group
    cmd += ' --num-cpu-threads %s' % self.soft.params['cpus']
    for p in ['presets']:
        if self.soft.params[p]:
            cmd += ' --%s %s' % (p.replace('_', '-'), self.soft.params[p])
    for p in [
        'mem_flag', 'bubble_level', 'prune_level', 'k_min', 'k_max', 'k_step',
        'min_count', 'prune_depth', 'cleaning_rounds', 'min_contig_len',
        'merge_level', 'disconnect_ratio', 'low_local_ratio', 'memory'
    ]:
        if self.soft.params[p]:
            cmd += ' --%s %s' % (p.replace('_', '-'), self.soft.params[p])
    for boolean in [
        'no_mercy', 'no_local', 'kmin_1pass', 'no_hw_accel', 'keep_tmp_files'
    ]:
        if self.soft.params[boolean]:
            cmd += ' --%s' % boolean.replace('_', '-')
    cmd += ' --out-dir %s' % out
    out_folder = '%s/intermediate_contigs' % out
    if self.soft.params['continue'] and not to_do(folder=out_folder):
        cmd += ' --continue'
    else:
        if len(inputs) == 3:
            cmd += ' --read %s -1 %s -2 %s' % tuple(inputs)
        if len(inputs) == 2:
            cmd += ' -1 %s -2 %s' % tuple(inputs)
        if len(inputs) == 1:
            cmd += ' --read %s' % inputs[0]
    cmd += '\nrm -rf %s\n' % tmp
    return cmd


def megahit(self) -> None:
    """MEGAHIT is an ultra-fast and memory-efficient NGS assembler. It is
    optimized for metagenomes, but also works well on generic single genome
    assembly (small or mammalian size) and single-cell assembly.

    References
    ----------
    Li, Dinghua, et al. "MEGAHIT: an ultra-fast single-node solution for
    large and complex metagenomics assembly via succinct de Bruijn graph."
    Bioinformatics 31.10 (2015): 1674-1676.

    Notes
    -----
    GitHub  : https://github.com/voutcn/megahit
    Docs    : https://github.com/voutcn/megahit/wiki
    Paper   : https://doi.org/10.1093/bioinformatics/btv033

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for MEGAHIT
        .pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    tech = 'illumina'
    for group, techs_inputs in self.inputs[self.sam_pool].items():

        if tech not in techs_inputs:
            sys.exit('[megahit] Illumina reads are required (skipped)')
        inputs = techs_inputs[tech]
        status_update(self, tech, inputs, group=group)

        tmp = '$TMPDIR/megahit_%s_%s' % (self.sam_pool, group)
        out = '%s/%s/%s' % (self.dir, self.sam_pool, group)
        self.outputs['dirs'].append(out)

        contigs = '%s/%s.contigs.fa' % (out, group)
        self.outputs['outs'][(tech, group)] = [contigs, out]

        if self.config.force or to_do(contigs):
            cmd = megahit_cmd(self, inputs, group, tmp, out)
            self.outputs['cmds'].setdefault(group, []).append(cmd)
            io_update(self, i_f=inputs, i_d=out, o_d=out, key=group)
            self.soft.add_status(tech, self.sam_pool, 1, group=group)
        else:
            self.soft.add_status(tech, self.sam_pool, 0, group=group)


def plass_cmd(
        self,
        fastqs: list,
        contigs: str,
        tmp_dir: str
) -> str:
    """Collect PLASS command.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    fastqs : list
        Paths to the fastqs input files
    contigs : str
        Path to the contigs output file
    tmp_dir : str
        Path to the temporary folder

    Returns
    -------
    cmd : str
        PLASS command
    """
    cmd = 'mkdir -p %s\n' % tmp_dir
    cmd += 'plass %s' % self.soft.params['type']
    cmd += ' %s' % ' '.join(fastqs)
    cmd += ' %s' % contigs
    cmd += ' %s' % tmp_dir
    cmd += ' --threads %s' % self.soft.params['cpus']
    for param in [
        'alph_size', 'k', 'split_memory_limit', 'e', 'c', 'a', 'cov_mode',
        'min_seq_id', 'min_aln_len', 'seq_id_mode', 'kmer_per_seq',
        'kmer_per_seq_scale', 'hash_shift', 'num_iterations', 'rescore_mode',
        'min_length', 'max_length', 'max_gaps', 'contig-start-mode',
        'contig_end_mode', 'orf_start_mode', 'forward_frames', 'reverse_frames',
        'translation_table', 'id_offset', 'protein_filter_threshold', 'dbtype',
        'db_load_mode', 'v', 'max_seq_len',
    ]:
        if len(param) == 1:
            cmd += ' -%s %s' % (param, self.soft.params[param])
        else:
            cmd += ' --%s %s' % (param, self.soft.params[param])

    for boolean in [
        'add_self_matches', 'spaced_kmer_mode', 'mask', 'mask_lower_case',
        'wrapped_scoring', 'adjust_kmer_len', 'include_only_extendable',
        'ignore_multi_kmer', 'translate', 'use_all_table_starts',
        'filter_proteins', 'shuffle', 'createdb_mode', 'compressed',
        'delete_tmp_inc', 'remove_tmp_files', 'filter_hits', 'sort_results',
        'create_lookup', 'write_lookup'
    ]:
        if self.soft.params[boolean]:
            cmd += ' --%s 1' % boolean
    cmd += '\n'
    cmd += 'rm -rf %s\n' % tmp_dir
    return cmd


def plass(self) -> None:
    """Plass (Protein-Level ASSembler) is a software to assemble short read
    sequencing data on a protein level. The main purpose of Plass is the
    assembly of complex metagenomic datasets. It assembles 10 times more
    protein residues in soil metagenomes than Megahit. Plass is GPL-licensed
    open source software that is implemented in C++ and available for Linux
    and macOS. The software is designed to run on multiple cores. Plass was
    used to create a Soil Reference Catalog (SRC) and a Marine Eukaryotic
    Reference Catalog (MERC).

    References
    ----------
    Steinegger, Martin, Milot Mirdita, and Johannes Söding. "Protein-level
    assembly increases protein sequence recovery from metagenomic samples
    manyfold." Nature methods 16.7 (2019): 603-606.

    Notes
    -----
    GitHub  : https://github.com/soedinglab/plass
    Docs    : https://ngs-docs.github.io/2018-cicese-metatranscriptomics/plass-paladin/
    Paper   : https://doi.org/10.1038/s41592-019-0437-4

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for Plass
        .sam : str
            Sample name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    for (tech, sam), fastqs in self.inputs[self.sam_pool].items():
        if tech != 'illumina':
            continue
        out = '/'.join([self.dir, tech, self.sam_pool])
        self.outputs['dirs'].append(out)
        status_update(self, tech, fastqs)

        if self.soft.params['type'] == 'nuclassemble':
            contigs = '%s/nucl_contigs.fasta' % out
        else:
            contigs = '%s/prot_contigs.fasta' % out
        self.outputs['outs'][(tech, sam)] = [contigs]

        if self.config.force or to_do(contigs):
            tmp_dir = '$TMPDIR/plass_%s' % self.sam_pool
            cmd = plass_cmd(self, fastqs, contigs, tmp_dir)
            self.outputs['cmds'].setdefault(tech, []).append(cmd)
            io_update(self, i_f=fastqs, o_d=out, key=tech)
            self.soft.add_status(tech, sam, 1)
        else:
            self.soft.add_status(tech, sam, 0)


def flye_cmd(
        self,
        tech: str,
        inputs: list,
        out: str
) -> str:
    """Create command lines for Flye.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    tech : str
        Technology: 'pacbio' or 'nanopore'
    inputs : list
        Path to the input files
    out : str
        Output folder for flye

    Returns
    -------
    cmd : str
        Flye command
    """
    params = tech_params(self, tech)
    cmd = 'flye'
    cmd += ' --%s %s' % (params['long_reads'], ' '.join(inputs))
    cmd += ' --out-dir %s' % out
    cmd += ' --threads %s' % params['cpus']
    if 'genome_size' in params:
        cmd += ' --genome-size %s' % params['genome_size']
    for boolean in ['meta', 'keep_haplotypes', 'scaffold', 'resume']:
        if params[boolean]:
            cmd += ' --%s' % boolean.replace('_', '-')
    for param in ['iterations', 'min_overlap', 'asm_coverage', 'read_error',
                  'polish_target', 'stop_after', 'resume_from']:
        if params[param]:
            cmd += ' --%s %s\n' % (param.replace('_', '-'), params[param])
    return cmd


def flye(self) -> None:
    """Flye is a de novo assembler for single-molecule sequencing reads,
    such as those produced by PacBio and Oxford Nanopore Technologies. It is
    designed for a wide range of datasets, from small bacterial projects to
    large mammalian-scale assemblies. The package represents a complete
    pipeline: it takes raw PacBio / ONT reads as input and outputs polished
    contigs. Flye also has a special mode for metagenome assembly.
    Currently, Flye will produce collapsed assemblies of diploid genomes,
    represented by a single mosaic haplotype. To recover two phased
    haplotypes consider applying HapDup after the assembly.

    References
    ----------
    Kolmogorov, Mikhail, et al. "metaFlye: scalable long-read metagenome
    assembly using repeat graphs." Nature Methods 17.11 (2020): 1103-1110.

    Notes
    -----
    GitHub  : https://github.com/fenderglass/Flye
    Paper   : https://doi.org/10.1038/s41592-020-00971-x

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for Flye
        .pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    for group, techs_inputs in self.inputs[self.sam_pool].items():

        if not sum([t in techs_inputs for t in ['pacbio', 'nanopore']]):
            sys.exit('[flye] Only for "pacbio" and/or "nanopore" (no input)')

        for tech, inputs in techs_inputs.items():
            if tech == 'illumina':
                continue

            out = '/'.join([self.dir, tech, self.sam_pool, group])
            self.outputs['dirs'].append(out)

            status_update(self, tech, inputs, group=group)

            contigs = '%s/assembly.fasta' % out
            gfa = '%s/assembly_graph.gfa' % out
            gv = '%s/assembly_graph.gv' % out
            info = '%s/assembly_info.txt' % out
            self.outputs['outs'][(tech, group)] = [contigs, info, gfa, gv]

            if self.config.force or to_do(contigs):
                cmd = flye_cmd(self, tech, inputs, out)
                tech_group = '%s_%s' % (tech, group)
                self.outputs['cmds'].setdefault(tech_group, []).append(cmd)
                io_update(self, i_f=inputs, i_d=out, o_d=out, key=tech_group)
                self.soft.add_status(tech, self.sam_pool, 1, group=group)
            else:
                self.soft.add_status(tech, self.sam_pool, 0, group=group)


def canu_cmd(
        self,
        tech: str,
        inputs: list,
        group: str,
        out: str
) -> str:
    """Collect CANU command.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    tech : str
        Technology: 'pacbio' or 'nanopore'
    inputs : list
        Input read fastx files
    group : str
        Name of the sample group within the pool
    out : str
        Output folder for CANU
    """
    params = tech_params(self, tech)
    cmd = 'if [ -d %s]; then rm -rf %s; fi\n' % (out, out)
    if 'path' in params:
        cmd += 'export PATH=$PATH:%s\n' % params['path']
    cmd += 'canu'
    cmd += ' -%s %s' % (tech, ' '.join(inputs))
    cmd += ' -d %s' % out
    cmd += ' -p %s' % group
    if 'specifications' in params:
        cmd += ' -s %s' % params['specifications']
    if 'genome_size' in params:
        cmd += ' genomeSize=%s' % params['genome_size']
    if params['processing']:
        cmd += ' -%s' % params['processing']
    if params['stage']:
        cmd += ' -%s' % params['stage']
    if params['rawErrorRate']:
        cmd += ' rawErrorRate=%s' % params['rawErrorRate']
    else:
        if tech == 'pacbio':
            cmd += ' rawErrorRate=0.300'
        else:
            cmd += ' rawErrorRate=0.500'
    if params['correctedErrorRate']:
        cmd += ' correctedErrorRate=%s' % params['correctedErrorRate']
    else:
        if tech == 'pacbio':
            cmd += ' correctedErrorRate=0.045'
        else:
            cmd += ' correctedErrorRate=0.144'

    for param in ['minReadLength', 'minOverlapLength', 'maxInputCoverage',
                  'corOutCoverage', 'corMhapSensitivity', 'corMinCoverage',
                  'redMemory', 'oeaMemory', 'batMemory']:
        cmd += ' %s=%s' % (param, params[param])

    for threads in ['bat', 'cns', 'cormhap', 'cormmap', 'corovl', 'cor',
                    'executive', 'hap', 'max', 'meryl', 'min', 'obtmhap',
                    'obtmmap', 'obtovl', 'oea', 'ovb', 'ovs', 'red',
                    'utgmhap', 'utgmmap', 'utgovl']:
        cmd += ' %sThreads=%s' % (threads, params['cpus'])
    return cmd


def canu(self) -> None:
    """Canu is a fork of the Celera Assembler designed for high-noise
    single-molecule sequencing (such as the PacBio RSII or Oxford Nanopore
    MinION).

    References
    ----------
    Koren, Sergey, et al. "Canu: scalable and accurate long-read assembly via
    adaptive k-mer weighting and repeat separation." Genome research 27.5
    (2017): 722-736.

    Notes
    -----
    GitHub  : https://github.com/marbl/canu
    Docs    : https://canu.readthedocs.io/en/latest/
    Paper   : https://doi.org/10.1101/gr.215087.116

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for CANU
        .pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    for group, techs_inputs in self.inputs[self.sam_pool].items():

        if not [t for t in ['pacbio', 'nanopore'] if techs_inputs.get(t)]:
            sys.exit('[canu] Only for "pacbio" and/or "nanopore" (no input)')

        for tech, inputs in techs_inputs.items():
            if tech == 'illumina':
                continue

            status_update(self, tech, inputs, group=group)

            out = '/'.join([self.dir, tech, self.sam_pool, group])
            self.outputs['dirs'].append(out)

            report = '%s/%s.report' % (out, group)
            contigs = '%s/%s.contigs.fasta' % (out, group)
            self.outputs['outs'][(tech, group)] = [contigs, report]

            if self.config.force or to_do(contigs):
                cmd = canu_cmd(self, tech, inputs, group, out)
                tech_group = '%s_%s' % (tech, group)
                self.outputs['cmds'].setdefault(tech_group, []).append(cmd)
                io_update(self, i_f=inputs, i_d=out, o_d=out, key=tech_group)
                self.soft.add_status(tech, self.sam_pool, 1, group=group)
            else:
                self.soft.add_status(tech, self.sam_pool, 0, group=group)


def unicycler_cmd(
        self,
        techs_inputs: dict,
        key: str,
        out: str,
        techs: list
) -> tuple:
    """Collect unicycler command.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    techs_inputs : dict
        Path to the input fastqs files per technology
    key : str
        Concatenation of the technology and co-assembly group
    out : str
        Output folder for unicycler
    techs : list
        Pair or technologies

    Returns
    -------
    cmd : str
        Unicycler commands
    inputs : list
        Path to all input fastqs files
    """
    tmp = '$TMPDIR/%s_%s_%s' % (self.soft.name, self.sam_pool, key)
    cmd = 'unicycler'
    inputs = []
    for tech in techs:
        if tech == 'illumina':
            illumina = tuple(techs_inputs[tech])
            inputs.extend(techs_inputs[tech])
            if len(illumina) == 1:
                cmd += ' --unpaired %s' % illumina[0]
            elif len(illumina) == 2:
                cmd += ' --short1 %s --short2 %s' % illumina
            elif len(illumina) == 3:
                cmd += ' --unpaired %s --short1 %s --short2 %s' % illumina
        else:
            cmd += ' --long %s' % techs_inputs[tech][0]
            inputs.extend(techs_inputs[tech])
    cmd += ' --out %s' % out
    cmd += ' --threads %s' % self.soft.params['cpus']
    cmd += ' --spades_tmp_dir %s' % tmp
    for param in [
        'contamination', 'bcftools_path', 'java_path', 'bowtie2_path',
        'bowtie2_build_path', 'samtools_path', 'pilon_path', 'scores',
        'start_genes', 'existing_long_read_assembly', 'tblastn_path',
        'makeblastdb_path', 'spades_path', 'racon_pth', 'min_bridge_qual',
        'start_gene_cov', 'start_gene_id', 'max_kmer_frac', 'min_kmer_frac',
        'depth_filter', 'mode', 'min_component_size', 'min_dead_end_size',
        'min_polish_size', 'min_anchor_seg_len', 'min_fasta_length', 'keep',
        'kmer_count', 'linear_seqs', 'low_score', 'kmers', 'verbosity'
    ]:
        if param in self.soft.params:
            cmd += ' --%s %s' % (param, self.soft.params[param])
    for boolean in ['vcf', 'no_correct', 'largest_component',
                    'no_miniasm', 'no_rotate', 'no_pilon']:
        if self.soft.params[boolean]:
            cmd += ' --%s' % self.soft.params[param]
    return cmd, inputs


def unicycler(self) -> None:
    """Unicycler is an assembly pipeline for bacterial genomes. It can
    assemble Illumina-only read sets where it functions as a
    SPAdes-optimiser. It can also assembly long-read-only sets (PacBio or
    Nanopore) where it runs a miniasm+Racon pipeline. For the best possible
    assemblies, give it both Illumina reads and long reads, and it will
    conduct a short-read-first hybrid assembly.

    References
    ----------
    Wick, Ryan R., et al. "Unicycler: resolving bacterial genome assemblies
    from short and long sequencing reads." PLoS computational biology 13.6
    (2017): e1005595.

    Notes
    -----
    GitHub  : https://github.com/rrwick/Unicycler
    Docs    : https://github.com/rrwick/Unicycler/wiki
    Paper   : https://doi.org/10.1371/journal.pcbi.1005595

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for unicycler
        .pool : str
            Pool name.
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    if self.sam_pool in self.pools:
        fastqs = self.inputs[self.sam_pool]
    else:
        fastqs = {self.sam_pool: self.inputs[self.sam_pool]}

    for sam_group, techs_inputs in fastqs.items():
        if 'illumina' not in techs_inputs:
            sys.exit('[unicycler] Illumina reads are required (skipped)')

        techs = ['illumina']
        if self.soft.params['hybrid']:
            if not techs_inputs.get(self.soft.params['hybrid']):
                print('[unicycler] No long reads/hybrid assembly:', sam_group)
            techs += [self.soft.params['hybrid']]

        hybrid = '_'.join(techs)
        tech_group = '%s_%s' % (hybrid, sam_group)
        if len(techs) > 1:
            hybrid = 'hybrid__%s' % hybrid

        out = '/'.join([self.dir, hybrid, self.sam_pool, sam_group])

        gfa = '%s/assembly.gfa' % out
        fasta = '%s/assembly.fasta' % out
        log = '%s/unicycler.log' % out
        self.outputs['outs'][(hybrid, sam_group)] = [fasta, gfa, log]

        if self.config.force or to_do(fasta):
            cmd, inputs = unicycler_cmd(
                self, techs_inputs, tech_group, out, techs)
            status_update(self, hybrid, inputs, group=sam_group)
            self.outputs['cmds'].setdefault(tech_group, []).append(cmd)
            io_update(self, i_f=inputs, o_d=out, key=tech_group)
            self.soft.add_status(hybrid, self.sam_pool, 1, group=sam_group)
        else:
            self.soft.add_status(hybrid, self.sam_pool, 0, group=sam_group)


# def miniasm_cmd(self):
#     cmd = '\nminimap2'
#     cmd += ' -x ava-pb'
#     cmd += ' -t8 pb-reads.fq pb-reads.fq'
#     cmd += ' | gzip -1 > reads.paf.gz'
#     cmd += ' > reads.paf.gz'
#     cmd += '\nminiasm'
#     cmd += ' -f reads.fq'
#     cmd += ' reads.paf.gz'
#     cmd += ' > reads.gfa'
#
#
# def miniasm(self):
#     cmd = miniasm_cmd(self)
#     pass


def necat_cmd(
        self,
        group: str,
        fastxs: list,
        out: str
) -> str:
    """Collect NECAT command.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    group : str
        Name of the sample group within the pool
    fastxs : list
        Paths to the input files
    out : str
        Output folder for NECAT

    Returns
    -------
    cmd : str
        NECAT command
    """
    reads_fp = '%s/reads.txt' % out
    cmd = necat_reads(fastxs, reads_fp)
    config_fp = '%s/config.txt' % out
    cmd += necat_config(self, reads_fp, config_fp, group)
    cmd += 'cd %s/..\n' % out
    cmd += 'export PATH="$PATH:%s"\n' % self.soft.params['path']
    cmd += 'necat.pl correct %s\n' % config_fp
    cmd += 'necat.pl assemble %s\n' % config_fp
    cmd += 'necat.pl bridge %s\n' % config_fp
    return cmd


def necat_reads(
        fastxs: list,
        reads_fp: str
) -> str:
    """Make the `reads.txt` input file.

    Parameters
    ----------
    fastxs : list
        Paths to the reads input files
    reads_fp : str
        Path to the `reads.txt` input file

    Returns
    -------
    cmd : str
        echo commands to make the `reads.txt` input file
    """
    cmd = ''
    for fdx, fastx in enumerate(fastxs):
        if fdx:
            cmd += 'echo -e "%s\\n" >> %s\n' % (fastx, reads_fp)
        else:
            cmd += 'echo -e "%s\\n" > %s\n' % (fastx, reads_fp)
    return cmd


def necat_config(
        self,
        reads_fp: str,
        config_fp: str,
        group: str
) -> str:
    """Create command lines for NECAT.

    Parameters
    ----------
    self : Commands class instance
        .soft.params
            Parameters
    reads_fp : str
        Path to the `reads.txt` input file
    config_fp : str
        Path to the `config.txt` input file
    group : str
        Name of the sample group within the pool

    Returns
    -------
    cmd : str
        echo commands to make the `config.txt` input file
    """

    opts = {x: [y for y in self.soft.params if y.startswith(y.rsplit('_')[0])]
            for x in self.soft.params if 'options' in x}

    config = list()
    config.append('PROJECT=%s' % group)
    config.append('ONT_READ_LIST=%s' % reads_fp)
    if self.soft.params['genome_size']:
        config.append('GENOME_SIZE=%s' % self.soft.params['genome_size'])
    config.append('THREADS=%s' % self.soft.params['cpus'])
    config.append('MIN_READ_LENGTH=%s' % self.soft.params['min_read_length'])
    config.append('PREP_OUTPUT_COVERAGE=%s' % self.soft.params[
        'prep_output_coverage'])
    for opt, opt_params in opts.items():
        params = ['-%s %s' % (x[-1], self.soft.params[x]) for x in opt_params]
        config.append('%s=%s' % (opt[:-2].upper(), ' '.join(params)))
    config.append('NUM_ITER=%s' % self.soft.params['num_iter'])
    config.append('CNS_OUTPUT_COVERAGE=%s' % self.soft.params[
        'cns_output_coverage'])
    config.append('CLEANUP=%s' % self.soft.params['cleanup'])
    config.append('POLISH_CONTIGS=%s' % self.soft.params['polish_contig'])

    cmd = ''
    for cdx, conf in enumerate(config):
        if cdx:
            cmd += 'echo -e "%s\\n" >> %s\n' % (conf, config_fp)
        else:
            cmd += 'echo -e "%s\\n" > %s\n' % (conf, config_fp)
    return cmd


def necat(self) -> None:
    """NECAT is an error correction and de-novo assembly tool for Nanopore
    long noisy reads.

    References
    ----------
    Chen, Ying, et al. "Efficient assembly of nanopore reads via highly
    accurate and intact error correction." Nature Communications 12.1
    (2021): 1-10.

    Notes
    -----
    GitHub  : https://github.com/xiaochuanle/necat
    Docs    : http://www.tgsbioinformatics.com/necat
    Paper   : https://doi.org/10.1038/s41467-020-20236-7

    Parameters
    ----------
    self : Commands class instance
        .dir : str
            Path to pipeline output folder for NECAT
        .pool : str
            Pool name
        .inputs : dict
            Input files
        .outputs : dict
            All outputs
        .soft.params
            Parameters
        .config
            Configurations
    """
    tech = 'nanopore'
    for group, fastxs in self.inputs[self.sam_pool].get(tech, {}).items():
        # define the output folder
        out = '%s/%s/%s' % (self.dir, self.sam_pool, group)

        status_update(self, tech, fastxs, group=group)

        # get some of the expected outputs files
        final = '%s/1-consensus/cns_final.fasta' % out
        contigs = '%s/4-fsa/contigs.fasta' % out
        bridged = '%s/6-bridge_contigs/bridged_contigs.fasta' % out
        polish = ''
        if self.soft.params['polish_contigs']:
            polish = '%s/6-bridge_contigs/polished_contigs.fasta' % out

        # [machinery] Collect the folder(s) to be created
        self.outputs['dirs'].append(out)
        # [machinery] Collect the outputs files in a standard / step-wise way
        self.outputs['outs'][(tech, group)] = [contigs, final, bridged, polish]

        # if user in command line want to force re-creation of tool's command
        # or if the file "contigs" is to do (i.e. not yet written by the tool)
        if self.config.force or to_do(contigs):
            cmd = necat_cmd(self, group, fastxs, out)
            # [machinery] Collect the command line
            self.outputs['cmds'].setdefault(group, []).append(cmd)
            io_update(self, i_f=fastxs, o_d=out, key=group)
            self.soft.add_status(tech, self.sam_pool, 1, group=group)
        else:
            self.soft.add_status(tech, self.sam_pool, 0, group=group)
